# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

import threading
from io import BytesIO
from time import monotonic
from typing import Any

try:
    from PIL import Image
except ImportError:
    Image = None

from xpra.os_util import gi_import
from xpra.util.io import load_binary_file
from xpra.net import compression
from xpra.util.str_fn import csv, memoryview_to_bytes
from xpra.util.env import envint, envbool
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("icon")

ARGB_ICONS = envbool("XPRA_ARGB_ICONS", True)
PNG_ICONS = envbool("XPRA_PNG_ICONS", True)
DEFAULT_ICONS = envbool("XPRA_DEFAULT_ICONS", True)

LOG_THEME_DEFAULT_ICONS = envbool("XPRA_LOG_THEME_DEFAULT_ICONS", False)
SAVE_WINDOW_ICONS = envbool("XPRA_SAVE_WINDOW_ICONS", False)
MAX_ARGB_PIXELS = envint("XPRA_MAX_ARGB_PIXELS", 1024)


class WindowIconSource:
    """
    Mixin for handling the sending of window icon pixels.
    """

    fallback_window_icon: bool | tuple[int, int, str, bytes] = False

    def __init__(self, window_icon_encodings, icons_encoding_options):
        self.window_icon_encodings = window_icon_encodings
        self.icons_encoding_options = icons_encoding_options  # icon caps

        self.has_png = PNG_ICONS and ("png" in self.window_icon_encodings)
        self.has_default = DEFAULT_ICONS and ("default" in self.window_icon_encodings)
        log("WindowIconSource(%s, %s) has_png=%s",
            window_icon_encodings, icons_encoding_options, self.has_png)

        self.window_icon_data = None
        self.send_window_icon_timer = 0
        self.theme_default_icons = icons_encoding_options.strtupleget("default.icons")
        self.window_icon_greedy = icons_encoding_options.boolget("greedy", False)
        self.window_icon_size = icons_encoding_options.intpair("size", (64, 64))
        self.window_icon_max_size = icons_encoding_options.intpair("max_size", self.window_icon_size)
        self.window_icon_max_size = (
            max(self.window_icon_max_size[0], 16),
            max(self.window_icon_max_size[1], 16),
        )
        self.window_icon_size = (
            min(self.window_icon_size[0], self.window_icon_max_size[0]),
            min(self.window_icon_size[1], self.window_icon_max_size[1]),
        )
        self.window_icon_size = (
            max(self.window_icon_size[0], 16),
            max(self.window_icon_size[1], 16),
        )
        log("client icon settings: size=%s, max_size=%s", self.window_icon_size, self.window_icon_max_size)
        if LOG_THEME_DEFAULT_ICONS:
            log("theme_default_icons=%s", self.theme_default_icons)

    def cleanup(self):
        self.cancel_window_icon_timer()

    def cancel_window_icon_timer(self):
        swit = self.send_window_icon_timer
        if swit:
            self.send_window_icon_timer = 0
            GLib.source_remove(swit)

    def get_info(self) -> dict[str, Any]:
        idata = self.window_icon_data
        if not idata:
            return {}
        w, h, fmt, data = idata
        return {
            "icon": {
                "width": w,
                "height": h,
                "format": fmt,
                "bytes": len(data),
            }
        }

    @staticmethod
    def get_fallback_window_icon():
        if WindowIconSource.fallback_window_icon is False and Image:
            try:
                from xpra.platform.paths import get_icon_filename  # pylint: disable=import-outside-toplevel
                icon_filename = get_icon_filename("xpra.png")
                log(f"get_fallback_window_icon() icon filename={icon_filename}")
                img = Image.open(icon_filename)
                icon_data = load_binary_file(icon_filename)
                icon = (img.size[0], img.size[1], "png", icon_data)
                WindowIconSource.fallback_window_icon = icon
                return icon
            except Exception as e:
                log.warn(f"Warning: failed to get fallback icon: {e}")
                WindowIconSource.fallback_window_icon = False
        return WindowIconSource.fallback_window_icon

    def get_default_window_icon(self, size=48):
        # return the icon which would be used from the wmclass
        wmclass_name = self.get_window_wm_class_name()
        if not wmclass_name:
            return None
        Gtk = gi_import("Gtk")
        it = Gtk.IconTheme.get_default()  # pylint: disable=no-member
        log("get_default_window_icon(%i) icon theme=%s, wmclass_name=%s", size, it, wmclass_name)
        for icon_name in (
                f"{wmclass_name}-color",
                wmclass_name,
                f"{wmclass_name}_{size}x{size}",
                f"application-x-{wmclass_name}",
                f"{wmclass_name}-symbolic",
                f"{wmclass_name}.symbolic",
        ):
            i = it.lookup_icon(icon_name, size, 0)
            log("lookup_icon(%s)=%s", icon_name, i)
            if not i:
                continue
            try:
                pixbuf = i.load_icon()
                log("load_icon()=%s", pixbuf)
                if pixbuf:
                    w, h = pixbuf.props.width, pixbuf.props.height
                    log("using '%s' pixbuf %ix%i", icon_name, w, h)
                    return w, h, "RGBA", pixbuf.get_pixels()
            except Exception:
                log("%s.load_icon()", i, exc_info=True)
        return None

    def get_window_wm_class_name(self):
        try:
            c_i = self.window.get_property("class-instance")
        except Exception:
            return None
        if not c_i or len(c_i) != 2:
            return None
        return c_i[0]

    def client_has_theme_icon(self):
        wm_class = self.get_window_wm_class_name()
        return wm_class and wm_class in self.theme_default_icons

    def send_window_icon(self):
        # some of this code could be moved to the work queue half, meh
        assert self.ui_thread == threading.current_thread()
        if self.suspended:
            return
        # this runs in the UI thread
        icons = self.window.get_property("icons")
        log("send_window_icon window %s found %i icons", self.window, len(icons or ()))
        if not icons:
            if self.client_has_theme_icon():
                log("%s in client theme icons already (not sending default icon)", self.theme_default_icons)
                return
            # try to load the icon for this class-instance from the theme:
            icons = []
            sizes = []
            for size in (self.window_icon_size, self.window_icon_max_size, (48, 64)):
                if size:
                    for dim in size:  # ie: 48
                        if dim not in sizes:
                            sizes.append(dim)
            for size in sizes:
                icon = self.window.get_default_window_icon(size)
                if icon:
                    icons.append(icon)
            if not icons:
                # try to find one using the wmclass:
                icon = self.get_default_window_icon()
                if icon:
                    log("send_window_icon window %s using default window icon", self.window)
                    icons.append(icon)
        max_w, max_h = self.window_icon_max_size
        icon = self.choose_icon(icons, max_w, max_h)
        if not icon:
            # try again, without size restrictions:
            # (we'll downscale it)
            icon = self.choose_icon(icons)
        if not icon:
            if not self.window_icon_greedy:
                return
            # "greedy": client does not set a default icon, so we must provide one every time
            # to make sure that the window icon does get set to something
            # (our icon is at least better than the window manager's default)
            if self.has_default:
                # client will set the default itself,
                # send a mostly empty packet:
                packet = ("window-icon", self.wid, 0, 0, "default", "")
                log("queuing window icon update: %s", packet)
                # this is cheap, so don't use the encode thread,
                # and make sure we don't send another one via the timer:
                self.cancel_window_icon_timer()
                self.queue_packet(packet, wait_for_more=True)
                return
            icon = WindowIconSource.get_fallback_window_icon()
            log("using fallback window icon")
        if not icon:
            log("no suitable icon")
            return
        self.window_icon_data = icon
        if not self.send_window_icon_timer:
            # call compress via the work queue
            # and delay sending it by a bit to allow basic icon batching:
            w, h = self.window_icon_data[:2]
            delay = min(1000, max(50, w * h * self.batch_config.delay_per_megapixel // 1000000))
            log("send_window_icon() window=%s, wid=%s, compression scheduled in %sms for batch delay=%i",
                self.window, self.wid, delay, self.batch_config.delay_per_megapixel)
            self.send_window_icon_timer = GLib.timeout_add(delay, self.call_in_encode_thread,
                                                           True, self.compress_and_send_window_icon)

    def compress_and_send_window_icon(self):
        # this runs in the work queue
        self.send_window_icon_timer = 0
        idata = self.window_icon_data
        if not idata or not self.has_png:
            return
        w, h, pixel_format, pixel_data = idata
        log("compress_and_send_window_icon() %ix%i in %s format, %i bytes for wid=%i",
            w, h, pixel_format, len(pixel_data), self.wid)
        if pixel_format not in ("BGRA", "RGBA", "png"):
            raise RuntimeError(f"invalid window icon format {pixel_format}")
        if pixel_format == "BGRA":
            # BGRA data is always unpremultiplied
            # (that's what we get from NetWMIcons)
            from xpra.codecs.argb.argb import premultiply_argb  # pylint: disable=import-outside-toplevel
            pixel_data = premultiply_argb(pixel_data)

        max_w, max_h = self.window_icon_max_size
        # use png if supported and if "premult_argb32" is not supported by the client (ie: html5)
        # or if we must downscale it (bigger than what the client is willing to deal with),
        # or if we want to save window icons
        must_scale = w > max_w or h > max_h
        log("compress_and_send_window_icon: %sx%s (max-size=%s, standard-size=%s), pixel_format=%s",
            w, h, self.window_icon_max_size, self.window_icon_size, pixel_format)
        must_convert = pixel_format != "png"
        log(" must convert=%s, must scale=%s", must_convert, must_scale)

        image = None
        if must_scale or must_convert or SAVE_WINDOW_ICONS:
            if Image is None:
                log("cannot scale or convert window icon without python-pillow")
                return
            # we're going to need a PIL Image:
            if pixel_format == "png":
                image = Image.open(BytesIO(pixel_data))
            else:
                image = Image.frombuffer("RGBA", (w, h), memoryview_to_bytes(pixel_data), "raw", pixel_format, 0, 1)
            if must_scale:
                # scale the icon down to the size the client wants
                # (we should scale + paste to preserve the aspect ratio, meh)
                icon_w, icon_h = self.window_icon_size
                if float(w) / icon_w >= float(h) / icon_h:
                    rh = min(max_h, h * icon_w // w)
                    rw = icon_w
                else:
                    rw = min(max_w, w * icon_h // h)
                    rh = icon_h
                log("scaling window icon down to %sx%s", rw, rh)
                try:
                    LANCZOS = Image.Resampling.LANCZOS
                except AttributeError:
                    LANCZOS = Image.LANCZOS
                image = image.resize((rw, rh), LANCZOS)
            if SAVE_WINDOW_ICONS:
                filename = f"server-window-{self.wid}-icon-{int(monotonic())}.png"
                image.save(filename, 'PNG')
                log("server window icon saved to %s", filename)

        if image:
            # image got converted or scaled, get the new pixel data:
            output = BytesIO()
            image.save(output, "png")
            pixel_data = output.getvalue()
            output.close()
            w, h = image.size
        wrapper = compression.Compressed("png", pixel_data)
        packet = ("window-icon", self.wid, w, h, wrapper.datatype, wrapper)
        log("queuing window icon update: %s", packet)
        self.queue_packet(packet, wait_for_more=True)

    @staticmethod
    def choose_icon(icons, max_w=1024, max_h=1024):
        if not icons:
            return None
        log("choose_icon from: %s", csv("%ix%i %s" % icon[:3] for icon in icons))
        size_image = {icon[0] * icon[1]: icon for icon in icons if icon[0] < max_w and icon[1] < max_h}
        if not size_image:
            return None
        # we should choose one whose size is close to what the client wants,
        # take the biggest one for now:
        largest_size = sorted(size_image)[-1]
        icon = size_image[largest_size]
        log("choose_icon(..)=%ix%i %s", *icon[:3])
        return icon



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.os_util import get_group_id, getuid, POSIX
from xpra.util.env import osexpand
from xpra.net.common import get_peercred
from xpra.net.bytestreams import SocketConnection
from xpra.scripts.config import TRUE_OPTIONS
from xpra.util.objects import typedict
from xpra.util.str_fn import csv


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log("peercred.Authenticator(%s)", kwargs)
        self.uid = -1
        self.gid = -1
        self.peercred_check = False
        if not POSIX:
            log.warn("Warning: peercred authentication is not supported on %r", os.name)
            return
        connection = kwargs.get("connection", None)
        uids = kwargs.pop("uid", "")
        gids = kwargs.pop("gid", "")
        allow_owner = kwargs.pop("allow-owner", "yes").lower() in TRUE_OPTIONS
        self.check_peercred(connection, uids, gids, allow_owner)
        super().__init__(**kwargs)

    def check_peercred(self, connection, uids="", gids="", allow_owner: bool = False) -> None:
        allow_uids = allow_gids = None
        if uids:
            allow_uids = []
            for x in uids.split(":"):
                if not x.strip():
                    continue
                x = osexpand(x.strip())
                try:
                    allow_uids.append(int(x))
                except ValueError:
                    import pwd  # pylint: disable=import-outside-toplevel
                    try:
                        pw = pwd.getpwnam(x)
                        allow_uids.append(pw.pw_uid)
                    except KeyError:
                        log.warn("Warning: unknown username '%s'", x)
            log("peercred: allow_uids(%s)=%s", uids, allow_uids)
        if gids:
            allow_gids = []
            for x in gids.split(":"):
                if not x.strip():
                    continue
                x = osexpand(x.strip())
                try:
                    allow_gids.append(int(x))
                except ValueError:
                    gid = get_group_id(x)
                    if gid >= 0:
                        allow_gids.append(gid)
                    else:
                        log.warn("Warning: unknown group '%s'", x)
            log("peercred: allow_gids(%s)=%s", gids, allow_gids)
        self.do_check_peercred(connection, allow_uids, allow_gids, allow_owner)

    def do_check_peercred(self, connection, allow_uids=None, allow_gids=None, allow_owner=False):
        try:
            if connection and isinstance(connection, SocketConnection):
                sock = connection._socket
                peercred = get_peercred(sock)
                log("get_peercred(%s)=%s", sock, peercred)
                if not peercred:
                    log.warn("Warning: failed to get peer credentials on %s", sock)
                    return
                _, uid, gid = peercred

                def check() -> bool:
                    if allow_owner and uid == getuid():
                        log(f"matched owner: {uid}")
                        return True
                    if allow_uids is not None and uid not in allow_uids:
                        log.warn("Warning: peercred access denied,")
                        log.warn(" uid %i is not in the whitelist: %s", uid, csv(allow_uids))
                        return False
                    if allow_gids is not None and gid not in allow_gids:
                        log.warn("Warning: peercred access denied,")
                        log.warn(" gid %i is not in the whitelist: %s", gid, csv(allow_gids))
                        return False
                    return True

                if check():
                    self.peercred_check = True
                    self.uid = uid
                    self.gid = gid
            else:
                log("peercred: invalid connection '%s' (not a socket connection)", connection)
        except Exception as e:
            log("peercred", exc_info=True)
            log.error("Error: cannot get peer uid")
            log.estr(e)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def requires_challenge(self) -> bool:
        return False

    def authenticate(self, _caps: typedict) -> bool:  # pylint: disable=arguments-differ
        return self.peercred_check

    def __repr__(self):
        return "peercred"



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position

import logging

from xpra.util.env import envbool
from xpra.log import Logger

PIL_DEBUG = envbool("XPRA_PIL_DEBUG", False)
if PIL_DEBUG:   # pragma: no cover
    log = Logger("encoder", "pillow")
    log.info("enabling PIL.DEBUG")
    level = logging.DEBUG
else:
    level = logging.INFO

# newer versions use this logger,
# we must initialize it before we load the class:
for x in ("Image", "PngImagePlugin", "WebPImagePlugin", "JpegImagePlugin"):
    logger = logging.getLogger("PIL.%s" % x)
    logger.setLevel(level)

import PIL  # noqa: E402
from PIL import Image  # noqa: E402
assert PIL is not None and Image is not None
Image.init()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.x11.gtk import gdk_display_source

gdk_display_source.init_gdk_display_source()

from xpra.x11.bindings.window import X11WindowBindings

X11Window = X11WindowBindings()

from xpra.x11.server.server_uuid import get_mode, get_uuid

print(f"mode={get_mode()}")
print(f"uuid={get_uuid()}")



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


# This class is used by the posix server to ensure
# we reap the dead pids so that they don't become zombies,
# also used for implementing --exit-with-children

import os
import signal
from typing import Any
from collections.abc import Callable, Sequence

from xpra.util.env import envint, envbool
from xpra.os_util import POSIX, gi_import
from xpra.log import Logger

GLib = gi_import("GLib")
log = Logger("server", "util", "exec")


class ProcInfo:
    __slots__ = (
        "pid", "pidfile", "pidinode",
        "name", "command", "ignore",
        "forget", "dead", "returncode",
        "callback", "process",
    )
    pid: int
    pidfile: str
    pidinode: int
    name: str
    command: Any
    ignore: bool
    forget: bool
    dead: bool
    returncode: int | None
    callback: Callable | None
    process: Any

    def __repr__(self):
        return f"ProcInfo({self.pid} : {self.command})"

    def get_info(self) -> dict[str, Any]:
        info = {
            "pid": self.pid,
            "name": self.name,
            "command": self.command,
            "ignore": self.ignore,
            "forget": self.forget,
            # not base types:
            # callback, process
            "dead": self.dead,
            "pidfile": self.pidfile,
            "pidinode": self.pidinode,
        }
        if self.returncode is not None:
            info["returncode"] = self.returncode
        return info


# Note that this class has async subtleties -- e.g., it is possible for a
# child to exit and us to receive the SIGCHLD before our fork() returns (and
# thus before we even know the pid of the child).  So be careful:
# We can also end up with multiple procinfo structures with the same pid,
# and that should be fine too
#
# WNOHANG is a tricky beast, see:
# https://github.com/gevent/gevent/issues/622
class ChildReaper:
    __slots__ = ("_quit", "_proc_info")

    # note: the quit callback will fire only once!

    def __init__(self, quit_cb=None):
        log("ChildReaper(%s)", quit_cb)
        self._quit = quit_cb
        self._proc_info = []
        USE_PROCESS_POLLING = not POSIX or envbool("XPRA_USE_PROCESS_POLLING")
        if USE_PROCESS_POLLING:
            POLL_DELAY = envint("XPRA_POLL_DELAY", 2)
            log("using process polling every %s seconds", POLL_DELAY)
            GLib.timeout_add(POLL_DELAY * 1000, self.check)
        else:
            signal.signal(signal.SIGCHLD, self.sigchld)

            # Check once after the mainloop is running, just in case the exit
            # conditions are satisfied before we even enter the main loop.
            # (Programming with unix the signal API sure is annoying.)

            def check_once():
                self.check()
                return False  # Only call once

            GLib.timeout_add(0, check_once)

    def cleanup(self) -> None:
        self.reap()
        self.poll()
        self._proc_info = []
        self._quit = None

    def add_process(self, process, name: str, command: str | Sequence[str], ignore=False, forget=False, callback=None) -> ProcInfo:
        pid = process.pid
        if pid <= 0:
            raise RuntimeError(f"process {process} has no pid!")
        procinfo = ProcInfo()
        procinfo.pid = pid
        procinfo.pidfile = ""
        procinfo.pidinode = 0
        procinfo.name = name
        procinfo.command = command
        procinfo.ignore = ignore
        procinfo.forget = forget
        procinfo.callback = callback
        procinfo.process = process
        procinfo.returncode = process.poll()
        procinfo.dead = procinfo.returncode is not None
        log("add_process%s pid=%s", (process, name, command, ignore, forget, callback), pid)
        # could have died already:
        self._proc_info.append(procinfo)
        if procinfo.dead:
            self.add_dead_process(procinfo)
        return procinfo

    def poll(self) -> bool:
        # poll each process that is not dead yet:
        log("poll() procinfo list: %s", self._proc_info)
        for procinfo in tuple(self._proc_info):
            process = procinfo.process
            if not procinfo.dead and process and process.poll() is not None:
                self.add_dead_process(procinfo)
        return True

    def set_quit_callback(self, cb: Callable) -> None:
        self._quit = cb

    def check(self) -> bool:
        # see if we are meant to exit-with-children
        # see if we still have procinfos alive (and not meant to be ignored)
        self.poll()
        watched = tuple(procinfo for procinfo in tuple(self._proc_info)
                        if not procinfo.ignore)
        alive = tuple(procinfo for procinfo in watched
                      if not procinfo.dead)
        cb = self._quit
        log("check() watched=%s, alive=%s, quit callback=%s", watched, alive, cb)
        if watched and not alive:
            if cb:
                self._quit = None
                cb()
            return False
        return True

    def sigchld(self, signum, frame) -> None:
        # we risk race conditions if doing anything in the signal handler,
        # better run in the main thread asap:
        GLib.idle_add(self._sigchld, signum, str(frame))

    def _sigchld(self, signum, frame_str) -> None:
        log("sigchld(%s, %s)", signum, frame_str)
        self.reap()

    def get_proc_info(self, pid: int) -> ProcInfo | None:
        for proc_info in tuple(self._proc_info):
            if proc_info.pid == pid:
                return proc_info
        return None

    def add_dead_pid(self, pid: int) -> None:
        # find the procinfo for this pid:
        matches = [procinfo for procinfo in self._proc_info if procinfo.pid == pid and not procinfo.dead]
        log("add_dead_pid(%s) matches=%s", pid, matches)
        if not matches:
            # not one of ours? odd.
            return
        for procinfo in matches:
            self.add_dead_process(procinfo)

    def add_dead_process(self, procinfo: ProcInfo) -> None:
        log("add_dead_process(%s)", procinfo)
        process = procinfo.process
        if procinfo.dead or not process:
            return
        procinfo.returncode = process.poll()
        procinfo.dead = procinfo.returncode is not None
        cb = procinfo.callback
        log("add_dead_process returncode=%s, dead=%s, callback=%s", procinfo.returncode, procinfo.dead, cb)
        if not procinfo.dead:
            log.warn("Warning: process '%s' is still running", procinfo.name)
            return
        if process and cb:
            procinfo.callback = None
            GLib.idle_add(cb, process)
        if procinfo.pidfile and procinfo.pidinode:
            from xpra.server.util import rm_pidfile
            rm_pidfile(procinfo.pidfile, procinfo.pidinode)
        # once it's dead, clear the reference to the process:
        # this should free up some resources
        # and also help to ensure we don't end up here again
        procinfo.process = None
        if procinfo.ignore:
            log("child '%s' with pid %s has terminated (ignored)", procinfo.name, procinfo.pid)
        else:
            log.info("child '%s' with pid %s has terminated", procinfo.name, procinfo.pid)
        if procinfo.forget:
            # forget it:
            try:
                self._proc_info.remove(procinfo)
            except ValueError:  # pragma: no cover
                log("failed to remove %s from proc info list", procinfo, exc_info=True)
        log("updated procinfo=%s", procinfo)
        self.check()

    def reap(self) -> None:
        self.poll()
        while POSIX:
            log("reap() calling os.waitpid%s", (-1, "WNOHANG"))
            try:
                pid = os.waitpid(-1, os.WNOHANG)[0]
            except OSError:
                break
            log("reap() waitpid=%s", pid)
            if pid == 0:
                break
            self.add_dead_pid(pid)

    def get_info(self) -> dict[Any, Any]:
        iv = tuple(self._proc_info)
        info: dict[Any, Any] = {
            "children": {
                "total": len(iv),
                "dead": len(tuple(True for x in iv if x.dead)),
                "ignored": len(tuple(True for x in iv if x.ignore)),
            }
        }
        pi = sorted(self._proc_info, key=lambda x: x.pid, reverse=True)
        cinfo: dict[int, Any] = info.setdefault("child", {})
        for i, procinfo in enumerate(pi):
            d = {}
            for k in ("name", "command", "ignore", "forget", "returncode", "dead", "pid"):
                v = getattr(procinfo, k)
                if v is None:
                    continue
                d[k] = v
            cinfo[i] = d
        return info


singleton: ChildReaper | None = None


def getChildReaper() -> ChildReaper:
    global singleton
    if singleton is None:
        singleton = ChildReaper()
    return singleton


def reaper_cleanup() -> None:
    s = singleton
    if s is not None:
        s.cleanup()
    # keep it around,
    # so we don't try to reinitialize it from the wrong thread
    # (signal requires the main thread)
    # singleton = None



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.server.mixins.servermixintest_util import ServerMixinTest
from unit.process_test_util import DisplayContext


class InputMixinTest(ServerMixinTest):

    def test_input(self):
        with DisplayContext():
            from xpra.server.mixins.input import InputServer
            from xpra.server.source.input import InputMixin
            opts = AdHocStruct()
            self._test_mixin_class(InputServer, opts, {}, InputMixin)

def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from OpenGL.GL import glEnable, GL_DONT_CARE, GL_TRUE

from xpra.util.env import envbool
from xpra.util.str_fn import strtobytes
from xpra.common import noop
from xpra.log import Logger

log = Logger("opengl")

OPENGL_DEBUG = envbool("XPRA_OPENGL_DEBUG", False)

GREMEDY_DEBUG = OPENGL_DEBUG
KHR_DEBUG = OPENGL_DEBUG

GL_DEBUG_OUTPUT: int = 0
GL_DEBUG_OUTPUT_SYNCHRONOUS: int = 0
gl_debug_callback: Callable = noop
glInitStringMarkerGREMEDY: Callable = noop
glStringMarkerGREMEDY: Callable = noop
glInitFrameTerminatorGREMEDY: Callable = noop
glFrameTerminatorGREMEDY: Callable = noop

if OPENGL_DEBUG:
    try:
        # pylint: disable=ungrouped-imports
        from OpenGL.GL import KHR

        GL_DEBUG_OUTPUT = int(KHR.debug.GL_DEBUG_OUTPUT)  # @UndefinedVariable
        GL_DEBUG_OUTPUT_SYNCHRONOUS = int(KHR.debug.GL_DEBUG_OUTPUT_SYNCHRONOUS)
        from OpenGL.GL.KHR.debug import glDebugMessageControl, glDebugMessageCallback, glInitDebugKHR
    except ImportError:
        log("Unable to import GL_KHR_debug OpenGL extension. Debug output will be more limited.")
        KHR_DEBUG = False
    try:
        from OpenGL.GL.GREMEDY import string_marker, frame_terminator

        glInitStringMarkerGREMEDY = string_marker.glInitStringMarkerGREMEDY
        glStringMarkerGREMEDY = string_marker.glStringMarkerGREMEDY
        glInitFrameTerminatorGREMEDY = frame_terminator.glInitFrameTerminatorGREMEDY
        glFrameTerminatorGREMEDY = frame_terminator.glFrameTerminatorGREMEDY
        from OpenGL.GL import GLDEBUGPROC

        def py_gl_debug_callback(source, error_type, error_id, severity, length, message, param) -> None:
            log.error("src %x type %x id %x severity %x length %d message %s, param=%s",
                      source, error_type, error_id, severity, length, message, param)

        gl_debug_callback = GLDEBUGPROC(py_gl_debug_callback)
        GREMEDY_DEBUG = all(bool(x) for x in (
            glInitStringMarkerGREMEDY,
            glStringMarkerGREMEDY,
            glInitFrameTerminatorGREMEDY,
            glFrameTerminatorGREMEDY,
        ))
    except ImportError:
        # This is normal- GREMEDY_string_marker is only available with OpenGL debuggers
        GREMEDY_DEBUG = False
        log("Unable to import GREMEDY OpenGL extension. Debug output will be more limited.")
log("OpenGL debugging settings:")
log(f" {GREMEDY_DEBUG=}")
log(f" {GL_DEBUG_OUTPUT=}, {GL_DEBUG_OUTPUT_SYNCHRONOUS}")
log(f" {gl_debug_callback=}")
log(f" {glInitStringMarkerGREMEDY=}, {glStringMarkerGREMEDY=}")
log(f" {glInitFrameTerminatorGREMEDY=}, {glFrameTerminatorGREMEDY=}")


def context_init_debug() -> None:
    global GREMEDY_DEBUG, KHR_DEBUG
    # Ask GL to send us all debug messages
    if KHR_DEBUG:
        if GL_DEBUG_OUTPUT and gl_debug_callback != noop and glInitDebugKHR() is True:
            glEnable(GL_DEBUG_OUTPUT)
            glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS)
            glDebugMessageCallback(gl_debug_callback, None)
            glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, None, GL_TRUE)
        else:
            KHR_DEBUG = False
    # Initialize string_marker GL debugging extension if available
    if GREMEDY_DEBUG:
        if glInitStringMarkerGREMEDY() is True:
            log.info("Extension GL_GREMEDY_string_marker available.")
            log.info(" Will output detailed information about each frame.")
        else:
            # General case - running without debugger, extension not available
            # don't bother trying again for another window:
            GREMEDY_DEBUG = False
        # Initialize frame_terminator GL debugging extension if available
        if glInitFrameTerminatorGREMEDY() is True:
            log.info("Enabling GL frame terminator debugging.")


def gl_marker(*msg) -> None:
    log(*msg)
    if not GREMEDY_DEBUG:
        return
    try:
        s = strtobytes(msg[0] % msg[1:])
    except TypeError:
        s = strtobytes(msg)
    from ctypes import c_char_p  # pylint: disable=import-outside-toplevel
    c_string = c_char_p(s)
    glStringMarkerGREMEDY(0, c_string)


def gl_frame_terminator() -> None:
    # Mark the end of the frame
    # This makes the debug output more readable especially when doing single-buffered rendering
    if not GREMEDY_DEBUG:
        return
    log("glFrameTerminatorGREMEDY()")
    glFrameTerminatorGREMEDY()



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, Gdk  # pylint: disable=wrong-import-position @UnresolvedImport


def main():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_size_request(220, 120)
    window.connect("delete_event", Gtk.main_quit)
    vbox = Gtk.VBox(homogeneous=False, spacing=0)

    def add_buttons(t1, cb1, t2, cb2):
        hbox = Gtk.HBox(homogeneous=True, spacing=10)
        b1 = Gtk.Button(t1)

        def vcb1(*_args):
            cb1()

        b1.connect('clicked', vcb1)
        hbox.pack_start(b1, expand=True, fill=False, padding=5)
        b2 = Gtk.Button(t2)

        def vcb2(*_args):
            cb2()

        b2.connect('clicked', vcb2)
        hbox.pack_start(b2, expand=True, fill=False, padding=5)
        vbox.pack_start(hbox, expand=False, fill=False, padding=2)

    def send_maximized_wm_state(mode):
        from xpra.x11.gtk.display_source import init_gdk_display_source
        init_gdk_display_source()
        from xpra.x11.bindings.window import constants, X11WindowBindings  #@UnresolvedImport
        X11Window = X11WindowBindings()
        root = window.get_window().get_screen().get_root_window()
        root_xid = root.get_xid()
        xwin = window.get_window().get_xid()
        SubstructureNotifyMask = constants["SubstructureNotifyMask"]
        SubstructureRedirectMask = constants["SubstructureRedirectMask"]
        event_mask = SubstructureNotifyMask | SubstructureRedirectMask
        X11Window.sendClientMessage(root_xid, xwin, False, event_mask, "_NET_WM_STATE", mode,
                                    "_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ", 0, 0)

    def maximize_X11(*args):
        send_maximized_wm_state(1)  #ADD

    def unmaximize_X11(*args):
        send_maximized_wm_state(2)  #REMOVE

    add_buttons("maximize", window.maximize, "unmaximize", window.unmaximize)
    add_buttons("maximize X11", maximize_X11, "unmaximize X11", unmaximize_X11)
    add_buttons("fullscreen", window.fullscreen, "unfullscreen", window.unfullscreen)

    def window_state(widget, event):
        STATES = {
            Gdk.WindowState.WITHDRAWN: "withdrawn",
            Gdk.WindowState.ICONIFIED: "iconified",
            Gdk.WindowState.MAXIMIZED: "maximized",
            Gdk.WindowState.STICKY: "sticky",
            Gdk.WindowState.FULLSCREEN: "fullscreen",
            Gdk.WindowState.ABOVE: "above",
            Gdk.WindowState.BELOW: "below",
        }
        print("window_state(%s, %s)" % (widget, event))
        print("flags: %s" % [STATES[x] for x in STATES.keys() if x & event.new_window_state])
    window.connect("window-state-event", window_state)

    window.add(vbox)
    window.show_all()
    Gtk.main()
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import dbus.service

from xpra.server.dbus.server import DBUS_Server, INTERFACE, ni
from xpra.log import Logger

log = Logger("dbus", "server")


class Shadow_DBUS_Server(DBUS_Server):

    @dbus.service.method(INTERFACE, in_signature='i')
    def SetRefreshDelay(self, milliseconds):
        log("SetRefreshDelay(%i)", milliseconds)
        return self.server.set_refresh_delay(ni(milliseconds))



#!/usr/bin/env python3

import sys
import json
from io import BytesIO
import os.path
from xpra.os_util import gi_import
from xpra.exit_codes import ExitValue

from PIL import Image
import cairo

Gtk = gi_import("Gtk")
Pango = gi_import("Pango")


winfo = "window.info"


def read_json(filename: str):
    with open(filename, "rb") as f:
        serialized = f.read()
    return json.loads(serialized)


def load_window_dir(dirpath: str) -> tuple[dict, list[dict]]:
    data = {}
    windowinfo_file = os.path.join(dirpath, winfo)
    if not os.path.exists(windowinfo_file):
        raise ValueError(f"directory {dirpath!r} does not look like a replay directory, {winfo!r} not found")
    windowinfo = read_json(windowinfo_file)

    for dirname in os.listdir(dirpath):
        if dirname == winfo:
            continue
        subdirpath = os.path.join(dirpath, dirname)
        if not os.path.isdir(subdirpath):
            print(f"warning: {subdirpath!r} is not a directory!")
            continue
        try:
            ts = int(dirname)
        except ValueError:
            raise ValueError(f"directory {dirname!r} is not a timestamp!") from None
        data[ts] = load_update_dir(subdirpath)
    updates = []
    for ts in sorted(data.keys()):
        updates.append(data[ts])
    return windowinfo, updates


def load_update_dir(dirpath: str) -> dict:
    data = {}
    for entry in os.listdir(dirpath):
        if not entry.endswith(".info"):
            continue
        flush = int(entry.split(".", 1)[0])
        filename = os.path.join(dirpath, entry)
        update_data = read_json(filename)
        update_data["directory"] = dirpath
        data[flush] = update_data
    return dict(reversed(sorted(data.items())))


class ReplayApp:

    def __init__(self, window_info: dict, replay_data: list[dict]):
        self.replay_data = replay_data
        self.frame = 0
        self.update = 0
        self.frame_data = {}
        self.backing = None
        self.window_info = window_info
        self.window_size = window_info.get("dimensions", (200, 200))
        self.drawing_area = Gtk.DrawingArea()
        self.drawing_area.connect("draw", self.draw_cb)
        self.replay_window = self.init_replay_window()
        self.control_window = self.init_control_window()
        self.draw_frame()

    def init_replay_window(self) -> Gtk.Window:
        replay_window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        replay_window.set_size_request(*self.window_size)
        replay_window.add(self.drawing_area)
        replay_window.connect("delete_event", Gtk.main_quit)
        replay_window.show_all()
        return replay_window

    def init_control_window(self) -> Gtk.Window:
        control_window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        control_window.set_transient_for(self.replay_window)
        control_window.set_size_request(1920, 720)
        control_window.connect("delete_event", Gtk.main_quit)
        vbox = Gtk.VBox(homogeneous=False, spacing=0)

        hbox = Gtk.HBox(homogeneous=True, spacing=10)
        vbox.pack_start(hbox, False, False, 10)

        restart = Gtk.Button(label="Restart")
        restart.connect("clicked", self.restart)
        hbox.add(restart)
        nf = Gtk.Button(label="Next Frame")
        nf.connect("clicked", self.next_frame)
        hbox.add(nf)

        hbox = Gtk.HBox(homogeneous=True, spacing=10)
        vbox.pack_start(hbox, False, False, 10)

        pu = Gtk.Button(label="Previous Update")
        pu.connect("clicked", self.previous_update)
        hbox.add(pu)
        nu = Gtk.Button(label="Next Update")
        nu.connect("clicked", self.next_update)
        hbox.add(nu)

        self.label = Gtk.Label()
        fontdesc = Pango.FontDescription("monospace 9")
        self.label.modify_font(fontdesc)
        self.label.set_margin_start(10)
        self.label.set_xalign(0)
        self.label.set_line_wrap(True)
        vbox.pack_start(self.label, True, True, 0)

        control_window.add(vbox)
        control_window.show_all()
        return control_window

    def update_label(self) -> None:
        frame_info = [
            f"frame {self.frame} / {len(self.replay_data)}",
            f"update {self.update} / {len(self.frame_data)}",
            "",
        ]

        for flush, data in self.frame_data.items():
            prefix = "****" if flush == self.update else "    "
            frame_info.append(f"{prefix}{flush} = {data}")
        self.label.set_label("\n".join(frame_info))

    def previous_update(self, *_args):
        if self.update > 0:
            self.update = self.update - 1
            self.draw_update()

    def next_update(self, *_args):
        self.update = self.update + 1
        if self.update >= len(self.frame_data):
            self.next_frame()
        else:
            self.draw_update()

    def draw_cb(self, _widget, cr):
        cr.set_source_surface(self.backing, 0, 0)
        cr.paint()
        return False

    def draw_update(self):
        self.update_label()
        update = self.frame_data[self.update]
        x = update["x"]
        y = update["y"]
        w = update["w"]
        h = update["h"]
        stride = update["stride"]
        options = update.get("options", {})
        if not options.get("paint", True):
            return

        ww, wh = self.window_size
        if self.backing is None:
            self.backing = cairo.ImageSurface(cairo.Format.ARGB32, ww, wh)

        encoding = update.get("encoding")
        filename = update.get("file", f"{self.update}.{encoding}")
        dirpath = update.get("directory")
        file_path = os.path.join(dirpath, filename)
        if encoding == "scroll":
            with open(file_path, "rb") as f:
                scrolls = json.loads(f.read())
            # ie: scrolls=[[0, 664, 1278, 103, 0, -264], [0, 1113, 1278, 116, 0, -264], ..]
            old_backing = self.backing
            self.backing = cairo.ImageSurface(cairo.Format.ARGB32, ww, wh)
            gc = cairo.Context(self.backing)
            gc.set_operator(cairo.Operator.SOURCE)
            gc.set_source_surface(old_backing, 0, 0)
            gc.rectangle(0, 0, ww, wh)
            gc.fill()
            for sx, sy, sw, sh, xdelta, ydelta in scrolls:
                gc.set_source_surface(old_backing, xdelta, ydelta)
                x = sx + xdelta
                y = sy + ydelta
                gc.rectangle(x, y, sw, sh)
                gc.fill()
            self.backing.flush()
        else:
            with open(file_path, "rb") as f:
                file_data = f.read()
            if encoding in ("rgb32", "rgb24"):
                rgb_data = file_data
                if options.get("lz4"):
                    from xpra.net.lz4.lz4 import decompress
                    rgb_data = decompress(file_data)
                rgb_format = options.get("rgb_format", "BGRX")
                if len(rgb_format) == 3:
                    from xpra.codecs.argb.argb import rgb_to_bgrx
                    rgb_data = rgb_to_bgrx(rgb_data)
            else:
                buf = BytesIO(file_data)
                img = Image.open(buf)
                img = img.convert("RGBA")
                rgb_data = img.tobytes("raw", img.mode)
            ba = bytearray(rgb_data)
            cfmt = cairo.FORMAT_ARGB32 if encoding != "rgb24" else cairo.FORMAT_RGB24
            surface = cairo.ImageSurface.create_for_data(ba, cfmt, w, h, stride or w*4)
            # paint in on the backing:
            ctx = cairo.Context(self.backing)
            ctx.set_source_surface(surface, x, y)
            ctx.paint()
        self.drawing_area.queue_draw()

    def restart(self, *_args) -> None:
        # we need to undo the updates!
        # (or just use the screenshot next?)
        self.frame = 0
        self.update = 0
        self.draw_frame()

    def next_frame(self, *_args) -> None:
        # make sure we apply all the updates before moving to the next frame:
        while self.update < len(self.frame_data) - 1:
            self.update = self.update + 1
            self.draw_update()
        self.frame = (self.frame + 1) % len(self.replay_data)
        self.update = 0
        self.draw_frame()

    def draw_frame(self, *_args) -> None:
        self.frame_data = self.replay_data[self.frame]
        self.draw_update()

    def run(self) -> ExitValue:
        Gtk.main()
        return 0


def main(argv: list[str]) -> int:
    if len(argv) != 2 or not os.path.isdir(argv[1]):
        print("usage: %s directory" % (argv[0]))
        return 1
    dirpath = argv[1]
    app = ReplayApp(*load_window_dir(dirpath))
    app.run()
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import os
import socket
import ipaddress
from queue import SimpleQueue
from typing import Union, cast
from collections.abc import Callable

from aioquic.quic.configuration import QuicConfiguration
from aioquic.quic.events import QuicEvent
from aioquic.quic.packet import QuicErrorCode
from aioquic.h3.connection import H3_ALPN
from aioquic.h0.connection import H0Connection
from aioquic.h3.connection import H3Connection
from aioquic.h3.events import (
    DataReceived,
    H3Event,
    HeadersReceived,
    PushPromiseReceived,
)
from aioquic.quic.connection import QuicConnection
from aioquic.asyncio.protocol import QuicConnectionProtocol

from xpra.os_util import POSIX
from xpra.scripts.config import InitExit
from xpra.exit_codes import ExitCode
from xpra.net.bytestreams import pretty_socket
from xpra.net.socket_util import create_udp_socket
from xpra.net.ssl_util import get_ssl_verify_mode
from xpra.net.quic.connection import XpraQuicConnection, aioquic_version_info
from xpra.net.quic.asyncio_thread import get_threaded_loop
from xpra.net.quic.common import USER_AGENT, MAX_DATAGRAM_FRAME_SIZE, binary_headers
from xpra.util.str_fn import csv, Ellipsizer
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("quic")

HttpConnection = Union[H0Connection, H3Connection]

IPV6 = socket.has_ipv6 and envbool("XPRA_IPV6", True)
PREFER_IPV6 = IPV6 and envbool("XPRA_PREFER_IPV6", POSIX)
HOSTS_PREFER_IPV4 = os.environ.get("XPRA_HOSTS_PREFER_IPV4", "localhost,127.0.0.1").split(",")
FAST_OPEN = envbool("XPRA_QUIC_FAST_OPEN", aioquic_version_info>=(1, 2))

WS_HEADERS: dict[str, str] = {
    ":method": "CONNECT",
    ":scheme": "https",
    ":protocol": "websocket",
    "sec-websocket-version": 13,
    "sec-websocket-protocol": "xpra",
    "user-agent": USER_AGENT,
}


class ClientWebSocketConnection(XpraQuicConnection):

    def __init__(self, connection: HttpConnection, stream_id: int, transmit: Callable[[], None],
                 host: str, port: int, info=None, options=None) -> None:
        super().__init__(connection, stream_id, transmit, host, port, "wss", info, options)
        self.write_buffer = SimpleQueue()

    def flush_writes(self) -> None:
        # flush the buffered writes:
        try:
            while self.write_buffer.qsize():
                self.stream_write(*self.write_buffer.get())
        finally:
            self.write_buffer = None

    def write(self, buf, packet_type: str = "") -> int:
        log(f"write(%s, %s) {len(buf)} bytes", Ellipsizer(buf), packet_type)
        if self.write_buffer is not None:
            # buffer it until we are connected and call flush_writes()
            self.write_buffer.put((buf, packet_type))
            return len(buf)
        return super().write(buf, packet_type)

    def http_event_received(self, event: H3Event) -> None:
        log("http_event_received(%s)", Ellipsizer(event))
        if isinstance(event, HeadersReceived):
            for header, value in event.headers:
                if header == b"sec-websocket-protocol":
                    subprotocols = value.decode().split(",")
                    if "xpra" not in subprotocols:
                        message = f"unsupported websocket subprotocols {subprotocols}"
                        log.warn(f"Warning: {message}")
                        if not self.accepted:
                            self.send_headers(self.stream_id, headers={":status": 501})
                            self.transmit()
                        self.close(QuicErrorCode.APPLICATION_ERROR, message)
                        return
                    self.accepted = True
                    self.flush_writes()
            return
        if isinstance(event, PushPromiseReceived):
            log(f"PushPromiseReceived: {event}")
            log(f"PushPromiseReceived headers: {event.headers}")
            return
        super().http_event_received(event)


class WebSocketClient(QuicConnectionProtocol):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._http: HttpConnection | None = None
        self._push_types: dict[str, int] = {}
        self._websockets: dict[int, ClientWebSocketConnection] = {}
        if self._quic.configuration.alpn_protocols[0].startswith("hq-"):
            self._http = H0Connection(self._quic)
        else:
            self._http = H3Connection(self._quic)

    def open(self, host: str, port: int, path: str) -> ClientWebSocketConnection:
        log(f"open({host}, {port}, {path})")
        stream_id = self._quic.get_next_available_stream_id()
        websocket = ClientWebSocketConnection(self._http, stream_id, self.transmit,
                                              host, port)
        self._websockets[stream_id] = websocket
        headers = {
            ":authority": host,
            ":path": path,
        }
        headers.update(WS_HEADERS)
        log("open: sending http headers for websocket upgrade")
        self._http.send_headers(stream_id=stream_id, headers=binary_headers(headers))
        self.transmit()
        return websocket

    def quic_event_received(self, event: QuicEvent) -> None:
        for http_event in self._http.handle_event(event):
            self.http_event_received(http_event)

    def http_event_received(self, event: H3Event) -> None:
        if not isinstance(event, (HeadersReceived, DataReceived, PushPromiseReceived)):
            log.warn(f"Warning: unexpected http event type: {event}")
            return
        stream_id = event.stream_id
        websocket: ClientWebSocketConnection | None = self._websockets.get(stream_id)
        if not websocket:
            # perhaps this is a new substream?
            sub = -1
            hdict = {}
            if isinstance(event, HeadersReceived):
                hdict = {k.decode(): v.decode() for k, v in event.headers}
                sub = int(hdict.get("substream", -1))
            if sub < 0:
                log.warn(f"Warning: unexpected websocket stream id: {stream_id} in {event}")
                return
            websocket = self._websockets.get(sub)
            if not websocket:
                log.warn(f"Warning: stream {sub} not found in {self._websockets}")
                return
            subtype = hdict.get("stream-type")
            log.info(f"new quic substream {stream_id} for {subtype} packets")
            self._websockets[stream_id] = websocket
        websocket.http_event_received(event)


def create_local_socket(family=socket.AF_INET) -> tuple[socket.socket, tuple[str, int]]:
    if family == socket.AF_INET6:
        local_host = "::"
    else:
        local_host = "0.0.0.0"
    local_port = 0
    sock = create_udp_socket(local_host, local_port, family)
    addr = (local_host, local_port)
    log(f"create_udp_socket({pretty_socket(addr)}, {family})={sock}")
    return sock, addr


async def get_address_options(host: str, port: int) -> tuple:
    if IPV6:
        family = socket.AF_UNSPEC
        family_options = (socket.AF_INET, socket.AF_INET6)
    else:
        family = socket.AF_INET
        family_options = (socket.AF_INET,)
    try:
        tl = get_threaded_loop()
        infos = await tl.loop.getaddrinfo(host, port, family=family, type=socket.SOCK_DGRAM)
        log(f"getaddrinfo({host}, {port}, {family}, SOCK_DGRAM)={infos}")
    except Exception as e:
        log(f"getaddrinfo({host}, {port}, {family}, SOCK_DGRAM)", exc_info=True)
        raise RuntimeError(f"cannot get address information for {pretty_socket((host, port))}: {e}") from None
    if IPV6:
        # we have to filter the results ourselves:
        infos = list(addr_info for addr_info in infos if addr_info[0] in family_options)
    has_ipv6 = any(addr_info[0] == socket.AF_INET6 for addr_info in infos)
    prefer_ipv4 = any(re.match(host_pattern, host) for host_pattern in HOSTS_PREFER_IPV4)
    if prefer_ipv4 and has_ipv6:
        # IPv6 is a waste of time on loopback,
        # and it can cause problems on some addresses,
        # so make sure it comes last in the list:
        infos.sort(key=lambda info: int(info[0] == socket.AF_INET6))
    elif PREFER_IPV6 and not has_ipv6:
        # no ipv6 returned, cook one up:
        ipv4_infos = tuple(addr_info for addr_info in infos if addr_info[0] == socket.AF_INET)
        # ie:( (<AddressFamily.AF_INET: 2>, <SocketKind.SOCK_DGRAM: 2>, 17, '', ('192.168.0.114', 10000)), )
        if ipv4_infos:
            ipv4 = ipv4_infos[0]
            addr = ipv4[4]  # ie: ('192.168.0.114', 10000)
            if len(addr) == 2:
                addr = ("::ffff:" + addr[0], addr[1], 0, 0)
                infos.insert(0, (socket.AF_INET6, socket.SOCK_DGRAM, ipv4[2], ipv4[3], addr))
                log(f"added IPv6 option: {infos[0]}")
    # ensure only the family_options we want are included,
    # (only really needed for AF_UNSPEC)
    return tuple(infos)


def quic_connect(host: str, port: int, path: str,
                 ssl_cert: str, ssl_key: str, ssl_key_password: str,
                 ssl_ca_certs, ssl_server_verify_mode: str, ssl_server_name: str):
    configuration = QuicConfiguration(
        alpn_protocols=H3_ALPN,
        is_client=True,
        max_datagram_frame_size=MAX_DATAGRAM_FRAME_SIZE,
    )
    configuration.verify_mode = get_ssl_verify_mode(ssl_server_verify_mode)
    if ssl_ca_certs:
        configuration.load_verify_locations(ssl_ca_certs)
    if ssl_cert:
        from pathlib import Path
        cert_path = Path(ssl_cert)
        key_path = Path(ssl_key) if ssl_key else None
        configuration.load_cert_chain(cert_path, key_path, ssl_key_password or None)
    if ssl_server_name:
        configuration.server_name = ssl_server_name
    else:
        # if host is not an IP address, use it for SNI:
        try:
            ipaddress.ip_address(host)
        except ValueError:
            configuration.server_name = host

    # configuration.max_data = args.max_data
    # configuration.max_stream_data = args.max_stream_data
    # configuration.quic_logger = QuicFileLogger(args.quic_log)
    # configuration.secrets_log_file = open(args.secrets_log, "a")

    def create_protocol() -> WebSocketClient:
        connection = QuicConnection(configuration=configuration)
        return WebSocketClient(connection)

    tl = get_threaded_loop()
    log(f"getting the list of address options for {host!r}:{port}")
    addresses = tl.sync(get_address_options, host, port)
    log(f"quic_connect will try {addresses=}")

    async def connect(addr_info):
        # ie:(AF_INET, SOCK_DGRAM, 0, '', ('192.168.0.10', 10000)
        log(f"connect({addr_info})")
        af = addr_info[0]
        sock, local_addr = create_local_socket(af)
        transport, protocol = await tl.loop.create_datagram_endpoint(create_protocol, sock=sock)
        log(f"{transport=}, {protocol=}")
        protocol = cast(QuicConnectionProtocol, protocol)
        addr = addr_info[4]  # ie: ('192.168.0.10', 10000)
        log(f"connecting from {pretty_socket(local_addr)} to {pretty_socket(addr)} with {FAST_OPEN=}")
        if FAST_OPEN:
            protocol.connect(addr, transmit=False)
        else:
            protocol.connect(addr)
        log("awaiting connected state")
        try:
            if not FAST_OPEN:
                await protocol.wait_connected()
            log(f"{protocol}.open({host!r}, {port}, {path!r})")
            conn = protocol.open(host, port, path)
            log(f"websocket connection {conn}")
            return conn
        except Exception as e:
            log("connect()", exc_info=True)
            # try to get a more meaningful exception message:
            einfo = str(e)
            if not einfo:
                quic_conn = getattr(protocol, "_quic", None)
                if quic_conn:
                    close_event = getattr(quic_conn, "_close_event", None)
                    log(f"{close_event=}, {dir(close_event)}")
                    if close_event:
                        err = close_event.error_code
                        msg = close_event.reason_phrase
                        if err & QuicErrorCode.CRYPTO_ERROR:
                            raise InitExit(ExitCode.CONNECTION_FAILED, msg)
                        # if (err & 0xFF)==QuicErrorCode.CONNECTION_REFUSED:
                        #    raise InitExit(ExitCode.CONNECTION_FAILED, msg)
                        raise RuntimeError(close_event.reason_phrase) from None
            raise RuntimeError(str(e)) from None

    # protocol.close()
    # await protocol.wait_closed()
    # transport.close()
    if len(addresses) == 1:
        return tl.sync(connect, addresses[0])

    errors = []
    for address in addresses:
        try:
            return tl.sync(connect, address)
        except Exception as ce:
            log("failed to connect:", exc_info=True)
            estr = str(ce) or type(ce)
            if estr not in errors:
                errors.append(estr)
    raise InitExit(ExitCode.CONNECTION_FAILED, "failed to connect: " + csv(errors))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

try:
    from xpra.rectangle import rectangle        #@UnresolvedImport

    R1 = rectangle(0, 0, 20, 20)
    R2 = rectangle(0, 0, 20, 20)
    R3 = rectangle(0, 0, 40, 40)
    R4 = rectangle(10, 10, 50, 50)
    R5 = rectangle(100, 100, 100, 100)
except ImportError:
    rectangle, R1, R2, R3, R4, R5 = None, None, None, None, None, None


# noinspection PyCallingNonCallable
class TestRegion(unittest.TestCase):

    def test_eq(self):
        assert R1==R2
        assert R1!=R3
        assert R2!=R3

    def test_intersection(self):
        r1 = rectangle(0, 0, 100, 100)
        r2 = rectangle(50, 50, 200, 200)
        i = r1.intersection_rect(r2)
        assert i.x==50 and i.y==50 and i.width==50 and i.height==50
        i2 = r2.intersection_rect(r1)
        assert i2==i
        r3 = rectangle(100, 100, 50, 50)
        i = r2.intersection_rect(r3)
        assert i==r3
        r4 = rectangle(0, 0, 10, 10)
        i = r3.intersection_rect(r4)
        assert i is None

    def test_contains(self):
        assert R1.contains_rect(R2)
        assert not R1.contains_rect(R3)
        assert R3.contains_rect(R1)
        assert not R1.contains_rect(R4)
        assert not R1.contains_rect(R5)

    def test_subtract(self):
        #  ##########          ##########
        #  #        #          ##########
        #  #        #          ##########
        #  #        #          ##########
        #  #   ##   #    ->                + ####  +  ####  +
        #  #   ##   #                        ####     ####
        #  #        #                                          ##########
        #  #        #                                          ##########
        #  #        #                                          ##########
        #  ##########                                          ##########
        r = rectangle(0, 0, 100, 100)
        sub = rectangle(40, 40, 20, 20)
        l = r.subtract_rect(sub)
        assert len(l)==4
        #verify total area has not changed:
        total = r.width*r.height
        assert total == sum(r.width*r.height for r in l+[sub])
        assert rectangle(0, 0, 100, 40) in l
        assert rectangle(0, 40, 40, 20) in l
        assert rectangle(0, 40, 40, 20) in l
        # at (0,0)
        # ##########
        # #        #
        # #        #
        # #        #
        # #        #         at (50, 50)
        # #        #         ##########
        # #        #         #        #
        # #        #    -    #        #
        # #        #         #        #
        # ##########         #        #
        #                    #        #
        #                    #        #
        #                    #        #
        #                    #        #
        #                    ##########
        r = rectangle(0, 0, 100, 100)
        sub = rectangle(50, 50, 100, 100)
        l = r.subtract_rect(sub)
        assert len(l)==2
        assert rectangle(0, 0, 100, 50) in l
        assert rectangle(0, 50, 50, 50) in l
        assert rectangle(200, 200, 0, 0) not in l


def main():
    #skip test if import failed (ie: not a server build)
    if rectangle is not None:
        unittest.main()
    else:
        print("no rectangle class, test skipped")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
import unittest

from xpra.os_util import gi_import
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.common import noop, SizedBuffer
from xpra.net.protocol import socket_handler, check
from xpra.net.protocol.constants import CONNECTION_LOST
from xpra.net.bytestreams import Connection
from xpra.net.compression import Compressed
from xpra.net.common import PacketType
from xpra.log import Logger

from unit.test_util import silence_error

log = socket_handler.log

GLib = gi_import("GLib")  # @UnresolvedImport

TIMEOUT = envint("XPRA_PROTOCOL_TEST_TIMEOUT", 20)
PROFILING = envbool("XPRA_PROTOCOL_PROFILING", False)
SHOW_PERF = envbool("XPRA_SHOW_PERF", False)


class FastMemoryConnection(Connection):
    def __init__(self, read_buffers, socktype="tcp"):
        self.read_buffers = read_buffers
        self.pos = 0
        self.write_data = []
        Connection.__init__(self, "local", socktype, {})

    def read(self, n) -> SizedBuffer:
        if self.read_buffers is None:
            while self.active:
                time.sleep(0.1)
            return b""
        if not self.read_buffers:
            logger = Logger("network")
            logger("read(%i) EOF", n)
            return b""
        b = self.read_buffers[0]
        if len(b) <= n:
            return self.read_buffers.pop(0)
        self.read_buffers[0] = b[n:]
        return b[:n]

    def write(self, buf, packet_type: str = "") -> int:
        self.write_data.append(buf)
        return len(buf)

    def __repr__(self):
        return "FastMemoryConnection"


def make_profiling_protocol_class(protocol_class):

    class ProfileProtocol(protocol_class):
        def profiling_context(self, basename):
            from pycallgraph import PyCallGraph, Config     #@UnresolvedImport
            from pycallgraph.output import GraphvizOutput   #@UnresolvedImport
            config = Config()
            graphviz = GraphvizOutput(output_file='%s-%i.png' % (basename, time.monotonic()))
            return PyCallGraph(output=graphviz, config=config)

        def write_format_thread_loop(self) -> None:
            with self.profiling_context("%s-format-thread" % protocol_class.TYPE):
                socket_handler.SocketProtocol.write_format_thread_loop(self)

        def do_read_parse_thread_loop(self) -> None:
            with self.profiling_context("%s-read-parse-thread" % protocol_class.TYPE):
                socket_handler.SocketProtocol.do_read_parse_thread_loop(self)

    return ProfileProtocol


class ProtocolTest(unittest.TestCase):
    protocol_class = socket_handler.SocketProtocol

    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        from xpra.net import packet_encoding
        packet_encoding.init_all()
        from xpra.net import compression
        compression.init_all()

    def make_memory_protocol(self, data=(b"", ), read_buffer_size=1, hangup_delay=0,
                             process_packet_cb=noop, get_packet_cb=socket_handler.no_packet):
        conn = FastMemoryConnection(data)
        if PROFILING:
            pc = make_profiling_protocol_class(self.protocol_class)
        else:
            pc = self.protocol_class
        p = pc(conn, process_packet_cb, get_packet_cb=get_packet_cb)
        p.read_buffer_size = read_buffer_size
        p.hangup_delay = hangup_delay
        assert p.get_info()
        assert repr(p)
        p.enable_default_compressor()
        p.enable_default_encoder()
        return p

    def test_verify_packet(self) -> None:
        verify_packet = check.verify_packet

        def nok(packet) -> None:
            assert verify_packet(packet) is False, f"packet {packet} should fail verification"

        def ok(packet) -> None:
            assert verify_packet(packet) is True, f"packet {packet} should not fail verification"

        # packets are iterable, so this should fail:
        for x in (True, 1, "hello", {}, None):
            nok(x)
        ok(("foo", 1))
        ok(("foo", [1,2,3], {1:2}))
        with silence_error(socket_handler):
            try:
                saved_verify_error_fn = check.verify_error
                check.verify_error = noop
                nok(["no-floats test", 1.1])
                nok(["no-None-values", [None], {1:2}])
                nok(["no-generic-objects", [1,2,3], {object() : 2}])
                nok(["no-nested-floats", [1,2,3], {1 : 2.2}])
            finally:
                check.verify_error = saved_verify_error_fn

    def test_invalid_data(self) -> None:
        self.do_test_invalid_data([b"\0"*1])
        self.do_test_invalid_data([b"P"*8])

    def do_test_invalid_data(self, data: SizedBuffer) -> None:
        errs = []
        proto = self.make_memory_protocol(data)

        def check_failed() -> None:
            if not proto.is_closed():
                errs.append("protocol not closed")
            if proto.input_packetcount > 0:
                errs.append("processed %i packets" % proto.input_packetcount)
            if proto.input_raw_packetcount == 0:
                errs.append("not read any raw packets")
            loop.quit()

        loop = GLib.MainLoop()
        GLib.timeout_add(500, check_failed)
        GLib.timeout_add(TIMEOUT*1000, loop.quit)
        proto.start()
        loop.run()
        assert not errs, csv(errs)

    def test_encoders_and_compressors(self) -> None:
        for encoder in ("rencodeplus", ):
            for compressor in ("lz4", ):
                p = self.make_memory_protocol()
                p.enable_encoder(encoder)
                p.enable_compressor(compressor)
                packet = ("test", 1, 2, 3)
                items = p.encode(packet)
                assert items

    def test_read_speed(self) -> None:
        if not SHOW_PERF:
            return
        total_size = 0
        total_elapsed = 0
        n_packets = 0
        for i in range(15, 19):
            n, size, elapsed = self.do_test_read_speed(2**i)
            total_size += size
            total_elapsed += elapsed
            n_packets += n
        print("%-9s incoming packet processing speed:\t%iMB/s" % (
            self.protocol_class.TYPE, total_size/total_elapsed//1024//1024)
        )
        print("%-9s packets parsed per second:\t\t%i" % (
            self.protocol_class.TYPE, n_packets/elapsed)
        )

    def do_test_read_speed(self, pixel_data_size=2**18, N=100) -> None:
        # prepare some packets to parse:
        p = self.make_memory_protocol()
        # use optimal setup:
        p.enable_encoder("rencodeplus")
        p.enable_compressor("lz4")
        # catch network packets before we write them:
        data = []

        def raw_write(_packet_type, items, *_args) -> None:
            for item in items:
                data.append(item)

        p.raw_write = raw_write
        packets = self.make_test_packets(pixel_data_size)
        for packet in packets:
            p._add_packet_to_queue(packet)
        ldata = self.repeat_list(data, N)
        total_size = sum(len(item) for item in ldata)
        # catch parsed packets:
        parsed_packets = []

        def process_packet_cb(proto, packet: PacketType) -> None:
            # log.info("process_packet_cb%s", packet[0])
            if packet[0] == CONNECTION_LOST:
                loop.quit()
            else:
                parsed_packets.append(packet[0])

        # run the protocol on this data:
        loop = GLib.MainLoop()
        GLib.timeout_add(TIMEOUT*1000, loop.quit)
        proto = self.make_memory_protocol(ldata, read_buffer_size=65536, process_packet_cb=process_packet_cb)
        start = time.monotonic()
        proto.start()
        loop.run()
        end = time.monotonic()
        assert len(parsed_packets)==N*3, "expected to parse %i packets but got %i" % (N*3, len(parsed_packets))
        elapsed = (end-start)
        log("do_test_read_speed(%i) %iMB in %ims", pixel_data_size, total_size, elapsed*1000)
        return N*len(packets), total_size, elapsed

    def make_test_packets(self, pixel_data_size=2**18) -> tuple:
        pixel_data = os.urandom(pixel_data_size)
        return (
            ("test", 1, 2, 3),
            ("ping", 100, 200, 300, 0),
            ("draw", 100, 100, 640, 480, Compressed("pixel-data", pixel_data), {}),
        )

    def repeat_list(self, items, N=100) -> list:
        #repeat the same pattern N times:
        l = []
        for _ in range(N):
            for item in items:
                assert item
                l.append(item)
        return l

    def test_format_thread(self) -> None:
        packets = self.make_test_packets()
        N = 10 if not SHOW_PERF else 1000
        many = self.repeat_list(packets, N)

        def get_packet_cb() -> tuple[PacketType, bool, bool]:
            try:
                packet = many.pop(0)
                return packet, False, True
            except IndexError:
                proto.close()
                return (), False, False

        def process_packet_cb(proto, packet: PacketType):
            if packet[0] == CONNECTION_LOST:
                GLib.timeout_add(1000, loop.quit)

        proto = self.make_memory_protocol(None, process_packet_cb=process_packet_cb, get_packet_cb=get_packet_cb)
        conn = proto._conn
        loop = GLib.MainLoop()
        GLib.timeout_add(TIMEOUT*1000, loop.quit)
        proto.enable_compressor("lz4")
        proto.enable_encoder("rencodeplus")
        proto.start()
        proto.source_has_more()
        start = time.monotonic()
        loop.run()
        end = time.monotonic()
        assert proto.is_closed()
        log("protocol: %s", socket_handler.SocketProtocol)
        log("%s write-data=%s", conn, len(conn.write_data))
        total_size = sum(len(packet) for packet in conn.write_data)
        elapsed = end-start
        log("bytes=%s, elapsed=%s", total_size, elapsed)
        if SHOW_PERF:
            print("\n")
            print("%-9s format thread:\t\t\t%iMB/s" % (proto.TYPE, int(total_size/elapsed//1024//1024)))
            n_packets = len(packets)*N
            print("%-9s packets formatted per second:\t\t%i" % (proto.TYPE, int(n_packets/elapsed)))
        assert conn.write_data


try:
    from xpra.net.websockets.protocol import WebSocketProtocol

    class WebsocketProtocolTest(ProtocolTest):
        protocol_class = WebSocketProtocol
except ImportError as e:
    log.warn("Warning: skipped websocket test")
    log.warn(" %s", e)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

_loop = None


def loop_init():
    global _loop
    if not _loop:
        from dbus.mainloop.glib import DBusGMainLoop, threads_init
        threads_init()
        _loop = DBusGMainLoop(set_as_default=True)
    return _loop


_session_bus = None


def init_session_bus(private=False):
    global _session_bus
    if _session_bus and not private:
        return _session_bus
    loop_init()
    import dbus
    _session_bus = dbus.SessionBus(private=private)
    return _session_bus


_system_bus = None


def init_system_bus():
    global _system_bus
    if _system_bus:
        return _system_bus
    loop_init()
    import dbus
    _system_bus = dbus.SystemBus()
    return _system_bus



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path

from xpra.util.io import get_util_logger
from xpra.util.env import envbool


def debug(*msg):
    """ delay import of logger to prevent cycles """
    log = get_util_logger()
    log.debug(*msg)
    return None


_gtkosx_warning_ = False


def do_get_resources_dir():
    rsc = None
    RESOURCES = "/Resources/"
    # FUGLY warning: importing gtkosx_application causes the dock to appear,
    # and in some cases we don't want that.. so use the env var XPRA_SKIP_UI as workaround for such cases:
    if not envbool("XPRA_SKIP_UI", False):
        from xpra.platform.darwin import get_OSXApplication
        macapp = get_OSXApplication()
        try:
            rsc = macapp.get_resource_path()
            debug("get_resources_dir() %s.get_resource_path=%s", macapp, rsc)
        except Exception:
            global _gtkosx_warning_
            if _gtkosx_warning_ is False:
                _gtkosx_warning_ = True
                # delayed import to prevent cycles:
                get_util_logger().error("Error: gtkosx_application module is missing - trying to continue anyway")
    else:
        debug("XPRA_SKIP_UI is set, not importing gtkosx_application")
    if rsc is None:
        # try using the path to this file to find the resource path:
        rsc = __file__
    i = rsc.rfind(RESOURCES)
    if i <= 0:
        # last fallback: try the default app dir
        from xpra.platform.paths import default_get_app_dir
        rsc = default_get_app_dir()
        debug("get_resources_dir() default_get_app_dir()=%s", rsc)
    i = rsc.rfind(RESOURCES)
    if i > 0:
        rsc = rsc[:i + len(RESOURCES)]
    debug("get_resources_dir()=%s", rsc)
    return rsc


def do_get_app_dir() -> str:
    from xpra.platform.paths import get_resources_dir
    rsc = get_resources_dir()
    CONTENTS = "/Contents/"
    i = rsc.rfind(CONTENTS)
    if i > 0:
        rsc = rsc[:i + len(CONTENTS)]
    debug("get_app_dir()=%s", rsc)
    return rsc  # hope for the best..


def do_get_image_dir() -> str:
    from xpra.platform.paths import get_resources_dir
    i = os.path.join(get_resources_dir(), "share", "xpra", "images")
    debug("get_image_dir()=%s", i)
    return i


def do_get_icon_dir() -> str:
    from xpra.platform.paths import get_resources_dir
    i = os.path.join(get_resources_dir(), "share", "xpra", "icons")
    debug("get_icon_dir()=%s", i)
    return i


def do_get_default_conf_dirs() -> list[str]:
    # the default config file we install into the Resources folder:
    # ie: /Volumes/Xpra/Xpra.app/Contents/Resources/etc
    from xpra.platform.paths import get_resources_dir
    return [os.path.join(get_resources_dir(), "etc", "xpra")]


def do_get_system_conf_dirs() -> list[str]:
    # the system-wide configuration directory
    dirs = []
    try:
        from Foundation import (
            NSSearchPathForDirectoriesInDomains,
            NSApplicationSupportDirectory,
            NSLocalDomainMask, NSSystemDomainMask
        )
        sdirs = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory,
                                                    NSLocalDomainMask | NSSystemDomainMask, False)
        for x in sdirs:
            # ie: "/Library/Application Support/Xpra"
            dirs.append(os.path.join(x, "Xpra"))
    except Exception:
        # fallback to hardcoded:
        default_conf_dir = "/Library/Application Support/Xpra"
        dirs = [os.environ.get("XPRA_SYSCONF_DIR", default_conf_dir)]
    dirs.append("/etc/xpra")
    return dirs


def do_get_ssh_conf_dirs() -> list[str]:
    return ["/private/etc", "/etc/ssh", "/usr/local/etc/ssh", "~/.ssh", "~/ssh"]


def do_get_user_conf_dirs(_uid) -> list[str]:
    # the system wide configuration directory
    dirs = []
    try:
        # when running sandboxed, it may look like this:
        # ~/Library/Containers/<bundle_id>/Data/Library/Application Support/
        from Foundation import (
            NSSearchPathForDirectoriesInDomains,
            NSApplicationSupportDirectory, NSUserDomainMask,
        )
        udirs = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, False)
        for x in udirs:
            dirs.append(os.path.join(x, "Xpra"))
    except Exception:
        # fallback to hardcoded:
        dirs = ["/Library/Application Support/Xpra"]
    dirs.append("~/.xpra")
    return dirs


def do_get_default_log_dirs() -> list[str]:
    dirs = []
    try:
        from Foundation import (
            NSSearchPathForDirectoriesInDomains,
            NSLibraryDirectory, NSUserDomainMask,
        )
        udirs = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, False)
        for x in udirs:
            # ie: ~/Library/
            dirs.append(os.path.join(x, "Logs", "Xpra"))
    except Exception:
        dirs.append("~/Library/Logs/Xpra")
    import tempfile
    dirs.append(tempfile.gettempdir())
    return dirs


def do_get_state_dir() -> str:
    return "~/.xpra/state"


def do_get_sessions_dir() -> str:
    return "~/.xpra/sessions"


def do_get_socket_dirs() -> list[str]:
    # return ["/var/tmp/%s-Xpra" % os.getuid(), "~/.xpra"]
    import tempfile
    return [tempfile.gettempdir(), "~/.xpra"]


def do_get_client_socket_dirs() -> list[str]:
    return ["~/.xpra/clients"]


def do_get_download_dir() -> str:
    d = "~/Downloads"
    try:
        from Foundation import (
            NSSearchPathForDirectoriesInDomains, NSDownloadsDirectory, NSUserDomainMask,
        )
        d = NSSearchPathForDirectoriesInDomains(NSDownloadsDirectory, NSUserDomainMask, False)[0]
        # (should be "~/Downloads")
    except Exception:
        pass
    if not os.path.exists(os.path.expanduser(d)):
        return "~"
    return d


def do_get_desktop_background_paths() -> list[str]:
    return [
        "/System/Library/CoreServices/DefaultDesktop.jpg",
        "/Library/Desktop Pictures/*jpg",
    ]


def do_get_sshpass_command() -> str:
    from xpra.platform.paths import get_app_dir
    base = get_app_dir()
    p = os.path.join(base, "Resources", "bin", "sshpass")
    if os.path.exists(p):
        return p
    return ""


def do_get_xpra_command() -> list[str]:
    if sys.argv and sys.argv[0].lower().endswith("/xpra"):
        return [sys.argv[0]]
    # try to use the one from the app bundle:
    from xpra.platform.paths import get_app_dir
    base = get_app_dir()
    xpra_cmd = os.path.join(base, "MacOS", "Xpra")
    if os.path.exists(xpra_cmd):
        return [xpra_cmd]
    return ["xpra"]


def _get_helpers_dir() -> str:
    from xpra.platform.paths import get_app_dir
    base = get_app_dir()
    p = os.path.join(base, "Helpers")
    return p


def do_get_nodock_command() -> list[str]:
    # try to use the sub-app:
    from xpra.platform.paths import get_app_dir
    base = get_app_dir()
    subapp = os.path.join(base, "Xpra_NoDock.app", "Contents")
    if os.path.exists(subapp) and os.path.isdir(subapp):
        base = subapp
    # AppStore builds have script wrappers:
    helper = os.path.join(base, "Resources", "scripts", "Xpra")
    if not os.path.exists(helper):
        helper = os.path.join(_get_helpers_dir(), "Xpra")
    if not os.path.exists(helper):
        # having a dock is still better than
        # trying to run a command that does not exist!
        from xpra.platform.paths import get_xpra_command
        return get_xpra_command()
    return [helper]


def do_get_audio_command() -> list[str]:
    return do_get_nodock_command()


def do_get_python_exec_command() -> list[str]:
    return [os.path.join(_get_helpers_dir(), "Python"), "-c"]


def do_get_python_execfile_command() -> list[str]:
    return [os.path.join(_get_helpers_dir(), "Python")]



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from unit.server_test_util import ServerTestUtil, log
from xpra.os_util import OSX, POSIX


class TestX11Keyboard(ServerTestUtil):

    @classmethod
    def setUpClass(cls):
        ServerTestUtil.setUpClass()
        display = cls.find_free_display()
        cls.xvfb = cls.start_Xvfb(display)
        os.environ["DISPLAY"] = display
        os.environ["GDK_BACKEND"] = "x11"
        from xpra.x11.bindings.posix_display_source import init_posix_display_source  #@UnresolvedImport
        cls.display_ptr = init_posix_display_source()
        from xpra.scripts.server import verify_gdk_display
        verify_gdk_display(display)

    @classmethod
    def tearDownClass(cls):
        from xpra.x11.bindings.posix_display_source import close_display_source  #@UnresolvedImport
        close_display_source(cls.display_ptr)
        ServerTestUtil.tearDownClass()
        cls.xvfb.terminate()

    def test_unicode(self):
        from xpra.x11.bindings.keyboard import X11KeyboardBindings  #@UnresolvedImport
        keyboard_bindings = X11KeyboardBindings()
        for x in (
                "2030", "0005", "0010", "220F", "2039", "2211",
                "2248", "FB01", "F8FF", "203A", "FB02", "02C6",
                "02DA", "02DC", "2206", "2044", "25CA",
        ):
            #hex form:
            hk = keyboard_bindings.parse_keysym("0x" + x)
            #osx U+ form:
            uk = keyboard_bindings.parse_keysym("U+" + x)
            log("keysym(U+%s)=%#x, keysym(0x%s)=%#x", x, uk, x, hk)
            assert hk and uk
            assert uk == hk, "failed to get unicode keysym %s" % x

    def test_grok_modifier_map(self):
        from xpra.x11.gtk.keys import grok_modifier_map
        grok_modifier_map(None)
        grok_modifier_map({})


def main():
    #can only work with an X11 server
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.client.gui.paint_colors import get_default_paint_box_color, get_paint_box_color, BLACK


class PaintColorsTest(unittest.TestCase):
	def test_defaults(self):
		assert get_default_paint_box_color()==BLACK
		assert get_paint_box_color("invalid-value")==BLACK
		for encoding in ("png", "h264"):
			assert get_paint_box_color(encoding)!=BLACK

def main():
	unittest.main()


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.os_util import POSIX
from xpra.platform.info import get_sys_info, get_version_info, get_user_info


class PlatformInfoTest(unittest.TestCase):

    def test_all_info(self):
        if POSIX:
            from xpra import common
            saved = common.FULL_INFO
            try:
                common.FULL_INFO = 2
                assert get_sys_info()
            finally:
                common.FULL_INFO = saved
        assert isinstance(get_version_info(), dict)
        assert isinstance(get_user_info(), dict)


def main():
    unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from time import monotonic, sleep
from typing import Any, Dict, Tuple
from collections.abc import Sequence
from ctypes import c_uint64, c_int, c_void_p, byref, POINTER
from ctypes import CDLL

from xpra.codecs.constants import VideoSpec, get_subsampling_divs, EncodingNotSupported
from xpra.codecs.image import ImageWrapper
from xpra.os_util import WIN32
from xpra.util.env import envbool, first_time
from xpra.util.objects import AtomicInteger, typedict
from xpra.log import Logger

from libc.stddef cimport wchar_t
from libc.stdint cimport uint8_t, int64_t, uintptr_t
from libc.string cimport memset, memcpy

from xpra.codecs.amf.amf cimport (
    amf_uint8,
    PLANE_TYPE_STR, SURFACE_FORMAT_STR, FRAME_TYPE_STR, MEMORY_TYPE_STR, DATA_TYPE_STR,
    AMF_FRAME_TYPE,
    AMF_RESULT, AMF_EOF, AMF_REPEAT,
    AMF_DX11_0,
    AMF_MEMORY_TYPE, AMF_MEMORY_DX11, AMF_MEMORY_HOST,
    AMF_INPUT_FULL, AMF_NOT_SUPPORTED, AMF_ALREADY_INITIALIZED, AMF_OK,
    AMF_SURFACE_FORMAT, AMF_SURFACE_YUV420P, AMF_SURFACE_NV12, AMF_SURFACE_BGRA,
    AMF_VARIANT_TYPE, AMF_VARIANT_INT64, AMF_VARIANT_SIZE, AMF_VARIANT_RATE,
    AMF_VIDEO_ENCODER_USAGE_ENUM,
    AMF_VIDEO_ENCODER_USAGE_LOW_LATENCY,
    AMF_VIDEO_ENCODER_USAGE_ULTRA_LOW_LATENCY,
    AMF_VIDEO_ENCODER_QUALITY_PRESET_ENUM,
    AMF_VIDEO_ENCODER_QUALITY_PRESET_QUALITY,
    AMF_VIDEO_ENCODER_QUALITY_PRESET_SPEED,
    AMF_VIDEO_ENCODER_QUALITY_PRESET_BALANCED,
    AMF_VIDEO_ENCODER_AV1_LEVEL_5_1,
    AMF_VIDEO_ENCODER_AV1_PROFILE_MAIN,
    AMF_VIDEO_ENCODER_AV1_USAGE_ULTRA_LOW_LATENCY,
    AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_ENUM,
    AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_HIGH_QUALITY,
    AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_QUALITY,
    AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_SPEED,
    AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_BALANCED,
    AMF_VIDEO_ENCODER_HEVC_USAGE_ENUM,
    AMF_VIDEO_ENCODER_HEVC_USAGE_ULTRA_LOW_LATENCY,
    AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_ENUM,
    AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_QUALITY,
    AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_SPEED,
    AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_BALANCED,
    amf_handle, amf_long, amf_int32,
    AMFDataAllocatorCB,
    AMFSize, AMFRate,
    AMFCaps,
    AMFGuid,
    AMFBuffer,
    AMFComponentOptimizationCallback,
    AMFPlane, AMFPlaneVtbl,
    AMFData, AMFDataVtbl,
    AMFTrace,
    AMFVariantStruct, AMFVariantInit,
    AMFVariantAssignInt64, AMFVariantAssignSize, AMFVariantAssignRate,
    AMFSurface, AMFSurfaceVtbl,
    AMFContext, AMFContext1,
    AMFComponent,
    AMFFactory,
)
from xpra.codecs.amf.common cimport (
    set_guid, get_factory, get_version, check, get_caps, get_plane_info, get_data_info, get_surface_info,
)


log = Logger("encoder", "amf")

DX11 = envbool("XPRA_AMF_DX11", WIN32)
SAVE_TO_FILE = os.environ.get("XPRA_SAVE_TO_FILE")


AMF_ENCODINGS : Dict[str, str] = {
    # AMFVideoEncoderVCE_SVC?
    "h264": "AMFVideoEncoderVCE_AVC",
    "hevc": "AMFVideoEncoder_HEVC",
    "av1": "AMFVideoEncoder_AV1",
}

START_TIME_PROPERTY = "StartTimeProperty"


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS
    # cdef extern from "unicodeobject.h":
    wchar_t* PyUnicode_AsWideCharString(object unicode, Py_ssize_t *size)
    # cdef extern from "pymem.h":
    void PyMem_Free(void *ptr)


cdef inline AMF_VIDEO_ENCODER_QUALITY_PRESET_ENUM get_h264_preset(int quality, int speed):
    if quality >= 80:
        return AMF_VIDEO_ENCODER_QUALITY_PRESET_QUALITY
    if speed >= 80:
        return AMF_VIDEO_ENCODER_QUALITY_PRESET_SPEED
    return AMF_VIDEO_ENCODER_QUALITY_PRESET_BALANCED


cdef inline AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_ENUM get_av1_preset(int quality, int speed):
    if quality >= 90:
        return AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_HIGH_QUALITY
    if quality >= 90:
        return AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_QUALITY
    if speed >= 80:
        return AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_SPEED
    return AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_BALANCED


cdef inline AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_ENUM get_hevc_preset(int quality, int speed):
    if quality >= 80:
        return AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_QUALITY
    if speed >= 80:
        return AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_SPEED
    return AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_BALANCED


def get_type() -> str:
    return "amf"


CODECS = tuple(AMF_ENCODINGS.keys())


def get_encodings() -> Sequence[str]:
    return CODECS


def get_info() -> Dict[str, Any]:
    info = {
        "version"       : get_version(),
        "encodings"     : get_encodings(),
    }
    return info


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    speed = 50
    quality = 50
    for encoding in CODECS:
        caps = CAPS.get(encoding, {})
        input_caps = caps.get("input", {})
        output_caps = caps.get("output", {})
        min_w, max_w = input_caps.get("width-range", (128, 4096))
        min_h, max_h = input_caps.get("height-range", (128, 4096))
        input_colorspaces = input_caps.get("native-formats", ("NV12", ))
        output_colorspaces = output_caps.get("native-formats", ("YUV420P", ))
        for input_colorspace in input_colorspaces:
            # ugly translation into the names xpra uses for legacy reasons:
            out_css = tuple({"NV12" : "YUV420P"}.get(cs, cs) for cs in output_colorspaces)
            spec = VideoSpec(
                encoding=encoding, input_colorspace=input_colorspace, output_colorspaces=out_css,
                has_lossless_mode=False,
                full_range=False,
                codec_class=Encoder, codec_type=get_type(),
                quality=quality, speed=speed,
                size_efficiency=60,
                setup_cost=20,
                min_w=min_w, min_h=min_h,
                max_w=max_w, max_h=max_h)
            log("AMF encoder spec for %r: %s", encoding, spec.to_dict())
            specs.append(spec)
    return specs


# these are the actual caps and are probed during the self-tests:
CAPS: Dict[str, str] = {}


ENCODING_CAPS: Dict[str, Dict[str, str]] = {
    "h264": {
        "max-bitrate" : "MaxBitrate",
        "number-of-streams": "NumOfStreams",
        "max-level": "MaxLevel",
        "b-frames": "BFrames",
        "max-throughput": "MaxThroughput",
        "fixed-sliced-mode": "FixedSliceMode",
        "color-conversion": "ColorConversion",
        "hardware-instances": "NumOfHwInstances",
        "requested-throughput": "RequestedThroughput",
        "roi": "ROIMap",
        "pre-analysis": "PreAnalysis",
    },
    "av1": {
        "max-bitrate" : "Av1MaxBitrate",
        # "number-of-streams": n/a
        "max-level": "Av1MaxLevel",
        "max-profile": "Av1MaxProfile",
        "hardware-instances": "Av1CapNumOfHwInstances",
        "max-throughput": "Av1CapMaxThroughput",
        "max-bitrate": "Av1MaxBitrate",
        "requested-throughput": "Av1CapRequestedThroughput",
        # "roi": n/a,
        "pre-analysis": "Av1PreAnalysis",
        "tile-output": "AV1SupportTileOutput",
        "width-alignment": "Av1WidthAlignmentFactor",
        "height-alignment": "Av1HeightAlignmentFactor",
    },
    "hevc": {
        "max-bitrate" : "HevcMaxBitrate",
        "number-of-streams": "HevcNumOfStreams",
        "max-profile": "HevcMaxProfile",
        "max-tier": "HevcMaxTier",
        "max-level": "HevcMaxLevel",
        "hardware-instances": "HevcNumOfHwInstances",
        "max-throughput": "HevcMaxThroughput",
        "requested-throughput": "HevcRequestedThroughput",
        "pre-analysis": "HevcPreAnalysis",
    },
}


generation = AtomicInteger()


cdef class Encoder:
    cdef AMFContext *context
    cdef AMFContext1 *context1
    cdef AMFComponent* encoder
    cdef AMF_SURFACE_FORMAT surface_format
    cdef AMFSurface* surface
    cdef void *device
    cdef object device_info
    cdef unsigned long frames
    cdef unsigned int width
    cdef unsigned int height
    cdef object encoding
    cdef object src_format
    cdef int speed
    cdef int quality
    cdef unsigned int generation
    cdef object file
    cdef object caps

    cdef object __weakref__

    def init_context(self, encoding: str,
                     unsigned int width, unsigned int height,
                     src_format: str,
                     options: typedict) -> None:
        log(f"amf init_context%s {SAVE_TO_FILE=}", (encoding, width, height, src_format, options))
        assert encoding in get_encodings(), "invalid encoding: %s" % encoding
        assert options.get("scaled-width", width)==width, "amf encoder does not handle scaling"
        assert options.get("scaled-height", height)==height, "amf encoder does not handle scaling"
        assert encoding in get_encodings()
        assert src_format == "NV12", f"invalid source format {src_format!r} for {encoding!r}"
        self.src_format = src_format
        if src_format == "YUV420P":
            self.surface_format = AMF_SURFACE_YUV420P
        elif src_format == "NV12":
            self.surface_format = AMF_SURFACE_NV12
        elif src_format == "BGRA":
            self.surface_format = AMF_SURFACE_BGRA
        else:
            raise ValueError("invalid pixel format {src_format!r}")

        self.encoding = encoding
        self.width = width
        self.height = height
        self.quality = options.intget("quality", 50)
        self.speed = options.intget("speed", 50)
        self.frames = 0

        self.context = NULL
        self.encoder = NULL
        self.device = NULL
        self.caps = {}
        self.device_info = {}

        self.amf_context_init()
        self.amf_encoder_init(options)

        self.generation = generation.increase()
        if SAVE_TO_FILE:
            filename = SAVE_TO_FILE+f"amf-{self.generation}.{encoding}"
            self.file = open(filename, "wb")
            log.info(f"saving {encoding} stream to {filename!r}")

    def check(self, res: AMF_RESULT, message: str):
        if res:
            check(res, message)
            self.clean()

    def amf_context_init(self) -> None:
        cdef AMFFactory *factory = get_factory()
        cdef AMF_RESULT res = factory.pVtbl.CreateContext(factory, &self.context)
        log(f"amf_context_init() CreateContext()={res}")
        self.check(res, "AMF context initialization")
        cdef AMFGuid amf1contextguid
        if DX11:
            res = self.context.pVtbl.InitDX11(self.context, NULL, AMF_DX11_0)
            log(f"amf_context_init() InitDX11()={res}")
            self.check(res, "AMF DX11 device initialization")
            self.device = self.context.pVtbl.GetDX11Device(self.context, AMF_DX11_0)
            assert self.device
            from xpra.platform.win32.d3d11.device import D3D11Device
            device = D3D11Device(<uintptr_t> self.device)
            self.device_info = device.get_info()
            log("device: %s", self.device_info)
            descr = self.device_info.get("description", "")
            if descr and first_time(f"GPU:{descr}"):
                log.info(f"AMF initialized using DX11 device {descr!r}")
        else:
            set_guid(&amf1contextguid, 0xd9e9f868, 0x6220, 0x44c6, 0xa2, 0x2f, 0x7c, 0xd6, 0xda, 0xc6, 0x86, 0x46)
            res = self.context.pVtbl.QueryInterface(self.context, &amf1contextguid, <void**> &self.context1)
            self.check(res, "AMFContext1 query")
            res = self.context1.pVtbl.InitVulkan(self.context1, NULL)
            log(f"amf_context_init() InitVulkan()={res}")
            if res != AMF_OK and res != AMF_ALREADY_INITIALIZED:
                self.check(res, "AMF Vulkan device initialization")
            self.device = <void *> self.context1.pVtbl.GetVulkanDevice(self.context1)
        log(f"amf_context_init() device=%#x", <uintptr_t> self.device)

    cdef void set_encoder_property(self, name: str, AMFVariantStruct var):
        cdef wchar_t *prop = PyUnicode_AsWideCharString(name, NULL)
        ret = self.encoder.pVtbl.SetProperty(self.encoder, prop, var)
        self.check(ret, f"AMF encoder setting property {name} to {var!r}")
        PyMem_Free(prop)

    def get_caps(self) -> Dict:
        assert self.encoder
        cdef AMFCaps *caps
        cdef AMF_RESULT r
        res = self.encoder.pVtbl.GetCaps(self.encoder, &caps)
        enc_props = ENCODING_CAPS.get(self.encoding, {})
        return get_caps(caps, enc_props)

    def amf_encoder_init(self, options: typedict) -> None:
        cdef AMFFactory *factory = get_factory()
        assert factory and self.context
        log(f"amf_encoder_init() for encoding {self.encoding!r}")

        amf_codec_name = str(AMF_ENCODINGS[self.encoding])
        cdef wchar_t *amf_codec = PyUnicode_AsWideCharString(amf_codec_name, NULL)
        cdef AMF_RESULT res = get_factory().pVtbl.CreateComponent(factory, self.context, amf_codec, &self.encoder)
        PyMem_Free(amf_codec)
        log(f"amf_encoder_init() CreateComponent()={res}")
        if res == AMF_NOT_SUPPORTED:
            message = f"{amf_codec_name!r} is not supported"
            log(message)
            raise EncodingNotSupported(message)
        self.check(res, f"AMF {self.encoding!r} encoder creation")

        self.caps = self.get_caps()
        CAPS[self.encoding] = self.caps
        log("encoder caps: %s", self.caps)

        speed = options.intget("speed", 50)
        quality = options.intget("quality", 50)
        bwlimit = options.intget("bandwidth-limit", 0)

        cdef AMFVariantStruct var
        cdef AMFSize size
        cdef AMFRate rate
        def setint64(prop: str, value: int) -> None:
            self.check(AMFVariantInit(&var), "AMF variant initialization")
            AMFVariantAssignInt64(&var, value)
            self.set_encoder_property(prop, var)
            # AMFVariantClear(&var)
        def setsize(prop: str) -> None:
            self.check(AMFVariantInit(&var), "AMF variant initialization")
            var.type = AMF_VARIANT_SIZE
            var.sizeValue.width = self.width
            var.sizeValue.height = self.height
            self.set_encoder_property(prop, var)
            # AMFVariantClear(&var)
        def setframerate(prop: str) -> None:
            self.check(AMFVariantInit(&var), "AMF variant initialization")
            var.type = AMF_VARIANT_RATE
            var.rateValue.num = 25
            var.rateValue.den = 1
            self.set_encoder_property(prop, var)
            # AMFVariantClear(&var)
        def setbitrate(prop: str) -> None:
            if bwlimit:
                setint64(prop, value=bwlimit)

        # tune encoder:
        if self.encoding == "h264":
            setsize("FrameSize")
            setframerate("FrameRate")
            setbitrate("TargetBitrate")
            # very unreliable way of detecting older cards
            # assume that newer ones have 4GB or more
            # (older cards may report 3.9GB)
            video_memory = self.device_info.get("memory", {}).get("video", 0)
            if video_memory >= 4*1024*1024*1024:
                setint64("Usage", value=AMF_VIDEO_ENCODER_USAGE_ULTRA_LOW_LATENCY)
            else:
                setint64("Usage", value=AMF_VIDEO_ENCODER_USAGE_LOW_LATENCY)
            setint64("QualityPreset", get_h264_preset(quality, speed))
            if False:
                setint64("Profile", value=AMF_VIDEO_ENCODER_PROFILE_HIGH)
                setint64("ProfileLevel", value=AMF_H264_LEVEL__5_1)
            if not options.boolget("b-frames", False):
                setint64("BPicturesPattern", value=0)
                # can be not supported - check Capability Manager sample
        elif self.encoding == "av1":
            setsize("Av1FrameSize")
            setframerate("Av1FrameRate")
            setbitrate("Av1TargetBitrate")
            setint64("Av1Usage", value=AMF_VIDEO_ENCODER_AV1_USAGE_ULTRA_LOW_LATENCY)
            setint64("Av1QualityPreset", get_av1_preset(quality, speed))
            setint64("Av1Profile", AMF_VIDEO_ENCODER_AV1_PROFILE_MAIN)
            setint64("Av1Level", AMF_VIDEO_ENCODER_AV1_LEVEL_5_1)
            if False:
                setint64("Av1AlignmentMode", AMF_VIDEO_ENCODER_AV1_ALIGNMENT_MODE_NO_RESTRICTIONS)
        elif self.encoding == "hevc":
            setint64("HevcUsage", AMF_VIDEO_ENCODER_HEVC_USAGE_ULTRA_LOW_LATENCY)
            setint64("HevcQualityPreset", get_hevc_preset(quality, speed))
            setsize("HevcFrameSize")
            setframerate("HevcFrameRate")
            setbitrate("HevcTargetBitrate")
            # setint64(AMF_VIDEO_ENCODER_HEVC_TIER, AMF_VIDEO_ENCODER_HEVC_TIER_HIGH)
            # setint64(AMF_VIDEO_ENCODER_HEVC_PROFILE_LEVEL, AMF_LEVEL_5_1)
        else:
            raise RuntimeError(f"unexpected encoding {self.encoding!r}")
        # init:
        res = self.encoder.pVtbl.Init(self.encoder, self.surface_format, self.width, self.height)
        self.check(res, f"AMF {self.encoding!r} encoder initialization for {self.width}x{self.height} {self.src_format}")
        log(f"amf_encoder_init() {self.encoding} encoder initialized at %#x", <uintptr_t> self.encoder)

    def is_ready(self) -> bool:
        return self.encoder != NULL

    def __repr__(self):
        return "amf.Encoder(%s)" % self.encoding

    def get_info(self) -> Dict[str, Any]:
        info = get_info()
        info |= {
            "frames"    : int(self.frames),
            "width"     : self.width,
            "height"    : self.height,
            "speed"     : self.speed,
            "quality"   : self.quality,
            "encoding"  : self.encoding,
            "src_format": self.src_format,
            "caps": self.caps,
        }
        return info

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def is_closed(self) -> bool:
        return self.context==NULL

    def get_type(self) -> str:
        return "amf"

    def get_src_format(self) -> str:
        return self.src_format

    def __dealloc__(self):
        self.clean()

    def clean(self) -> None:
        log("clean() encoder=%s", <uintptr_t> self.encoder)
        cdef AMFComponent* encoder = self.encoder
        if encoder:
            self.encoder = NULL
            encoder.pVtbl.Terminate(encoder)
            encoder.pVtbl.Release(encoder)
        log("clean() context=%s", <uintptr_t> self.context)
        cdef AMFContext *context = self.context
        if context:
            self.context = NULL
            context.pVtbl.Terminate(context)
            context.pVtbl.Release(context)
        self.context1 = NULL
        self.frames = 0
        self.width = 0
        self.height = 0
        self.encoding = ""
        self.src_format = ""
        f = self.file
        if f:
            self.file = None
            f.close()

    def compress_image(self, image: ImageWrapper, options: typedict) -> Tuple[bytes, Dict]:
        cdef uint8_t *pic_in[2]
        cdef int strides[2]
        assert self.context!=NULL
        pixels = image.get_pixels()
        istrides = image.get_rowstride()
        pf = image.get_pixel_format().replace("A", "X")
        if pf != self.src_format:
            raise ValueError(f"expected {self.src_format} but got {image.get_pixel_format()}")
        assert image.get_width()==self.width, "invalid image width %s, expected %s" % (image.get_width(), self.width)
        assert image.get_height()==self.height, "invalid image height %s, expected %s" % (image.get_height(), self.height)
        assert pixels, "failed to get pixels from %s" % image
        assert len(pixels)==2, "image pixels does not have 2 planes! (found %s)" % len(pixels)
        assert len(istrides)==2, "image strides does not have 2 values! (found %s)" % len(istrides)
        divs = get_subsampling_divs(self.src_format)

        cdef int speed = options.intget("speed", 50)
        if speed>=0:
            self.set_encoding_speed(speed)
        cdef int quality = options.intget("quality", 50)
        if quality>=0:
            self.set_encoding_quality(quality)

        cdef Py_buffer py_buf[2]
        for i in range(2):
            memset(&py_buf[i], 0, sizeof(Py_buffer))
        try:
            for i in range(2):
                xdiv, ydiv = divs[i]
                if PyObject_GetBuffer(pixels[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                    raise ValueError(f"failed to read pixel data from {type(pixels[i])}")
                assert istrides[i]>=self.width//xdiv, "invalid stride %i for width %i" % (istrides[i], self.width)
                assert py_buf[i].len>=istrides[i]*(self.height//ydiv), "invalid buffer length %i for plane %s, at least %i needed" % (
                    py_buf[i].len, "YUV"[i], istrides[i]*(self.height//ydiv))
                pic_in[i] = <uint8_t *> py_buf[i].buf
                strides[i] = istrides[i]
            return self.do_compress_image(pic_in, strides), {
                "csc"       : {"NV12": "YUV420P"}.get(self.src_format, self.src_format),
                "frame"     : int(self.frames),
                "full-range" : False,
                #"quality"  : min(99+self.lossless, self.quality),
                #"speed"    : self.speed,
            }
        finally:
            self.frames += 1
            for i in range(2):
                if py_buf[i].buf:
                    PyBuffer_Release(&py_buf[i])

    cdef void alloc_surface(self, AMFSurface **surface, AMF_MEMORY_TYPE memory=AMF_MEMORY_HOST):
        assert self.context
        memtype = MEMORY_TYPE_STR(memory)
        cdef AMF_RESULT res = self.context.pVtbl.AllocSurface(self.context, memory, self.surface_format,
                                                             self.width, self.height, surface)
        if res:
            self.check(res, f"AMF {memtype} surface initialization for {self.width}x{self.height} {self.src_format}")
        cdef AMFSurface *ptr = surface[0]
        assert ptr != NULL
        if log.is_debug_enabled():
            log(f"{memtype} surface: %s", get_surface_info(ptr))

    cdef void set_surface_property(self, AMFSurface *surface, name: str, variant: AMF_VARIANT_TYPE, value: int64_t):
        cdef AMFVariantStruct var
        AMFVariantAssignInt64(&var, value)
        cdef wchar_t *prop = PyUnicode_AsWideCharString(name, NULL)
        surface.pVtbl.SetProperty(surface, prop, var)
        PyMem_Free(prop)

    cdef uintptr_t get_native_plane(self, AMFSurface *surface, unsigned int plane_index):
        # get the D3D11 host destination surface pointer for this plane:
        plane = surface.pVtbl.GetPlaneAt(surface, plane_index)
        assert plane
        if log.is_debug_enabled():
            log("plane=%s", get_plane_info(plane))
        # texture is a `ID3D11Texture2D`:
        cdef uintptr_t texture = <uintptr_t> plane.pVtbl.GetNative(plane)
        assert texture
        return texture

    cdef bytes do_compress_image(self, uint8_t *pic_in[2], int strides[2]):
        cdef unsigned long start_time = 0    # nanoseconds!
        cdef AMFPlane *plane
        cdef uintptr_t src, dst
        cdef uintptr_t host_texture
        cdef uintptr_t gpu_texture
        cdef AMFSurface *host_surface
        cdef AMFSurface *gpu_surface
        cdef amf_int32 hpitch, vpitch
        cdef int stride
        cdef AMF_RESULT res

        if not WIN32:
            raise ImportError("amf encoder needs porting to this platform")

        from xpra.platform.win32.d3d11.device import D3D11Device
        log("D3D11Device(%#x)", <uintptr_t> self.device)
        device = D3D11Device(<uintptr_t> self.device)
        with device.get_device_context() as dc:
            if log.is_debug_enabled():
                log("device: %s", device.get_info())
                log("device context: %s", dc.get_info())

            self.alloc_surface(&host_surface, AMF_MEMORY_HOST)

            for plane_index in range(2):
                log("plane %s src data=%#x", ["Y", "UV"][plane_index], <uintptr_t> pic_in[plane_index])
                plane = host_surface.pVtbl.GetPlaneAt(host_surface, plane_index)
                assert plane
                stride = strides[plane_index]
                hpitch = plane.pVtbl.GetHPitch(plane)
                vpitch = plane.pVtbl.GetVPitch(plane)
                host_texture = <uintptr_t> plane.pVtbl.GetNative(plane)
                src = <uintptr_t> pic_in[plane_index]
                dst = <uintptr_t> host_texture
                if hpitch == stride:
                    # same stride, happy path with just one big copy:
                    memcpy(<void *> dst, <void *> src, hpitch * vpitch)
                else:
                    for y in range(vpitch):
                        memcpy(<void *> dst, <void *> src, stride)
                        dst += hpitch
                        src += stride

            # make it accessible by the GPU:
            res = host_surface.pVtbl.Convert(host_surface, AMF_MEMORY_DX11)
            self.check(res, "AMF Convert to DX11 memory")

            self.alloc_surface(&gpu_surface, AMF_MEMORY_DX11)

            for plane_index in range(2):
                log("plane %s", ["Y", "UV"][plane_index])
                host_texture = self.get_native_plane(host_surface, plane_index)
                gpu_texture = self.get_native_plane(gpu_surface, plane_index)
                dc.copy_resource(gpu_texture, host_texture)
            log("flush()")
            dc.flush()
            log("freeing host surface=%s", <uintptr_t> host_surface)
            host_surface.pVtbl.Release(host_surface)

        ns = round(1000 * 1000 * monotonic())
        self.set_surface_property(gpu_surface, START_TIME_PROPERTY, AMF_VARIANT_INT64, ns)

        log("encoder.SubmitInput()")
        res = self.encoder.pVtbl.SubmitInput(self.encoder, <AMFData*> gpu_surface)
        gpu_surface.pVtbl.Release(gpu_surface)
        if res == AMF_INPUT_FULL:
            raise RuntimeError("AMF encoder input is full!")
        self.check(res, "AMF submitting input to the encoder")
        cdef AMFData* data = NULL
        for i in range(200):
            res = self.encoder.pVtbl.QueryOutput(self.encoder, &data)
            log(f"QueryOutput()={res}, data=%#x", <uintptr_t> data)
            if res == 0:
                break
            if res == AMF_EOF:
                return b""
            if res == AMF_REPEAT:
                sleep(0.001)
                continue
            self.check(res, "AMF query output")
        assert data
        if log.is_debug_enabled():
            log("data=%s", get_data_info(data))
        cdef AMFGuid guid
        set_guid(&guid, 0xb04b7248, 0xb6f0, 0x4321, 0xb6, 0x91, 0xba, 0xa4, 0x74, 0xf, 0x9f, 0xcb)
        cdef AMFBuffer* buffer = NULL
        cdef uint8_t *output = NULL
        cdef size_t size = 0
        try:
            res = data.pVtbl.QueryInterface(data, &guid, <void**> &buffer)
            log(f"QueryInterface()={res} AMFBuffer=%#x", <uintptr_t> buffer)
            self.check(res, "AMF data query interface")
            assert buffer != NULL
            output = <uint8_t*> buffer.pVtbl.GetNative(buffer)
            size = buffer.pVtbl.GetSize(buffer)
            log(f"output=%#x, size=%i", <uintptr_t> output, size)
            assert output and size
            bdata = output[:size]
            if self.file:
                self.file.write(bdata)
            return bdata
        finally:
            if buffer != NULL:
                buffer.pVtbl.Release(buffer)
            data.pVtbl.Release(data)

    def flush(self, unsigned long frame_no) -> None:
        cdef AMF_RESULT res
        for _ in range(200):
            res = self.encoder.pVtbl.Drain(self.encoder)
            if res != AMF_INPUT_FULL:
                break
            sleep(0.001)
        self.check(res, "AMF encoder flush")

    def set_encoding_speed(self, int pct) -> None:
        if self.speed==pct:
            return
        self.speed = pct
        self.do_set_encoding_speed(pct)

    cdef void do_set_encoding_speed(self, int speed):
        pass

    def set_encoding_quality(self, int pct) -> None:
        if self.quality==pct:
            return
        self.quality = pct
        self.do_set_encoding_quality(pct)

    cdef void do_set_encoding_quality(self, int pct):
        pass


def selftest(full=False) -> None:
    global CODECS, SAVE_TO_FILE
    from xpra.codecs.checks import testencoder, get_encoder_max_size
    from xpra.codecs.amf import encoder
    temp = SAVE_TO_FILE
    try:
        SAVE_TO_FILE = None
        CODECS = testencoder(encoder, full)
        log("AMF specs()=%s", get_specs())
    finally:
        SAVE_TO_FILE = temp



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from time import monotonic
from typing import NoReturn
from collections import namedtuple
from collections.abc import Iterable, Callable

from xpra.gstreamer.common import import_gst, format_element_options
from xpra.audio.gstreamer_util import (
    parse_audio_source, get_source_plugins, get_sink_plugins, get_default_sink_plugin, get_default_source,
    can_decode, can_encode, get_muxers, get_demuxers, get_all_plugin_names,
)
from xpra.net.subprocess_wrapper import SubprocessCaller, SubprocessCallee, exec_kwargs, exec_env
from xpra.platform.paths import get_audio_command
from xpra.common import FULL_INFO
from xpra.os_util import WIN32, OSX, POSIX, BITS, gi_import
from xpra.util.parsing import parse_simple_dict
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.scripts.config import InitExit, InitException
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("audio")

DEBUG_SOUND = envbool("XPRA_SOUND_DEBUG", False)
SUBPROCESS_DEBUG = tuple(x.strip() for x in os.environ.get("XPRA_SOUND_SUBPROCESS_DEBUG", "").split(",") if x.strip())
FAKE_START_FAILURE = envbool("XPRA_SOUND_FAKE_START_FAILURE", False)
FAKE_EXIT = envbool("XPRA_SOUND_FAKE_EXIT", False)
FAKE_CRASH = envbool("XPRA_SOUND_FAKE_CRASH", False)
SOUND_START_TIMEOUT = envint("XPRA_SOUND_START_TIMEOUT", 5000 * (1 + int(WIN32)))

DEFAULT_SOUND_COMMAND_ARGS = os.environ.get("XPRA_DEFAULT_SOUND_COMMAND_ARGS", "--windows=no").split(",")


def get_full_audio_command() -> list[str]:
    return get_audio_command() + DEFAULT_SOUND_COMMAND_ARGS


def get_audio_wrapper_env() -> dict[str, str]:
    env = {
        "XPRA_LOG_SOCKET_STATS": os.environ.get("XPRA_LOG_SOCKET_STATS", "0"),
    }
    if WIN32:
        # we don't want the output to go to a log file
        env["XPRA_REDIRECT_OUTPUT"] = "0"
    elif POSIX and not OSX:
        try:
            from xpra.audio.pulseaudio.util import add_audio_tagging_env
            add_audio_tagging_env(env)
        except ImportError as e:
            log.warn("Warning: failed to set pulseaudio tagging:")
            log.warn(" %s", e)
    return env


# this wrapper takes care of launching src.py or sink.py
#
# the command line should look something like:
# xpra MODE IN OUT PLUGIN PLUGIN_OPTIONS CODECS CODEC_OPTIONS VOLUME
# * MODE can be _audio_record or _audio_play
# * IN is where we read the encoded commands from, specify "-" for stdin
# * OUT is where we write the encoded output stream, specify "-" for stdout
# * PLUGIN is the audio source (for recording) or sink (for playing) to use, can be omitted (will be auto detected)
#   ie: pulsesrc, autoaudiosink
# * PLUGIN_OPTIONS is a string containing options specific to this plugin
#   ie: device=somedevice,otherparam=somevalue
# * CODECS:  the list of codecs that we are willing to support
#   ie: mp3,flac
# * CODECS_OPTIONS: a string containing options to apply to the codec
#   ie: blocksize=1024,otherparam=othervalue
# * VOLUME: optional, a number from 0.0 to 1.0
#   ie: 1.0
# FIXME: CODEC_OPTIONS should allow us to specify different options for each CODEC
# The output will be a regular xpra packet, containing serialized signals that we receive
# The input can be a regular xpra packet, those are converted into method calls

class AudioSubprocess(SubprocessCallee):
    """ Utility superclass for audio subprocess wrappers
        (see audio_record and audio_play below)
    """

    def __init__(self, wrapped_object, method_whitelist, exports_list):
        # add bits common to both record and play:
        methods = method_whitelist + ["set_volume", "cleanup"]
        exports = ["state-changed", "info", "error"] + exports_list
        super().__init__(wrapped_object=wrapped_object, method_whitelist=methods)
        for x in exports:
            self.connect_export(x)

    def start(self) -> None:
        if not FAKE_START_FAILURE:
            def wrap_start() -> bool:
                self.wrapped_object.start()
                return False
            GLib.idle_add(wrap_start)
        if FAKE_EXIT > 0:
            def process_exit() -> None:
                self.cleanup()
                GLib.timeout_add(250, self.stop)

            GLib.timeout_add(FAKE_EXIT * 1000, process_exit)
        if FAKE_CRASH > 0:
            def force_exit() -> NoReturn:
                sys.exit(1)

            GLib.timeout_add(FAKE_CRASH * 1000, force_exit)
        super().start()

    def cleanup(self) -> None:
        wo = self.wrapped_object
        log("cleanup() wrapped object=%s", wo)
        if wo:
            # this will stop the audio pipeline:
            self.wrapped_object = None
            try:
                wo.cleanup()
            except RuntimeError:
                log("cleanup() failed to clean %s", wo, exc_info=True)
        GLib.timeout_add(1000, self.do_stop)

    def export_info(self) -> bool:
        wo = self.wrapped_object
        if wo:
            self.send("info", wo.get_info())
        return wo is not None


class AudioRecord(AudioSubprocess):
    """ wraps SoundSource as a subprocess """

    def __init__(self, *pipeline_args):
        from xpra.audio.src import AudioSource
        audio_pipeline = AudioSource(*pipeline_args)
        super().__init__(audio_pipeline, [], ["new-stream", "new-buffer"])
        self.large_packets = ["new-buffer"]


class AudioPlay(AudioSubprocess):
    """ wraps AudioSink as a subprocess """

    def __init__(self, *pipeline_args):
        from xpra.audio.sink import AudioSink
        audio_pipeline = AudioSink(*pipeline_args)
        super().__init__(audio_pipeline, ["add_data"], [])


def run_audio(mode: str, error_cb: Callable, options, args) -> int:
    """ this function just parses command line arguments to feed into the audio subprocess class,
        which in turn just feeds them into the audio pipeline class (sink.py or src.py)
    """
    gst = import_gst()
    if not gst:
        return 1
    info = mode.replace("_audio_", "")  # ie: "_audio_record" -> "record"
    from xpra.platform import program_context
    with program_context(f"Xpra-Audio-{info}", f"Xpra Audio {info}"):
        log("run_audio(%s, %s, %s, %s) gst=%s", mode, error_cb, options, args, gst)
        if info == "record":
            subproc = AudioRecord
        elif info == "play":
            subproc = AudioPlay
        elif info == "query":
            plugins = get_all_plugin_names()
            sources = [x for x in get_source_plugins() if x in plugins]
            sinks = [x for x in get_sink_plugins() if x in plugins]
            from xpra.audio.gstreamer_util import get_gst_version
            d = {
                "encoders": can_encode(),
                "decoders": can_decode(),
                "sources": sources,
                "source.default": get_default_source() or "",
                "sinks": sinks,
                "sink.default": get_default_sink_plugin() or "",
                "muxers": get_muxers(),
                "demuxers": get_demuxers(),
            }
            if FULL_INFO >= 1:
                d["gst.version"] = tuple(int(x) for x in get_gst_version())
            if FULL_INFO >= 2:
                d |= {
                    "plugins": plugins,
                    "python.version": sys.version_info[:3],
                    "python.bits": BITS,
                }
            for k, v in d.items():
                if isinstance(v, (list, tuple)):
                    v = ",".join(str(x) for x in v)
                print(f"{k}={v}")
            return 0
        else:
            log.error(f"Error: unknown mode {mode!r}")
            return 1
        assert len(args) >= 6, "not enough arguments"

        # the plugin to use (ie: 'pulsesrc' for src.py or 'autoaudiosink' for sink.py)
        plugin = args[2]
        # plugin options (ie: "device=monitor_device,something=value")
        options = parse_simple_dict(args[3])
        # codecs:
        codecs = [x.strip() for x in args[4].split(",")]
        # codec options:
        codec_options = parse_simple_dict(args[5])
        # volume (optional):
        try:
            volume = int(args[6])
        except (ValueError, IndexError):
            volume = 1.0

        ss = None
        try:
            ss = subproc(plugin, options, codecs, codec_options, volume)
            ss.start()
            return 0
        except InitExit as e:
            log.error(f"{info}: {e}")
            return e.status
        except InitException as e:
            log.error(f"{info}: {e}")
            return 1
        except OSError:
            log.error("run_audio%s error", (mode, error_cb, options, args), exc_info=True)
            return 1
        finally:
            if ss:
                ss.stop()


def _add_debug_args(command: list[str]) -> None:
    from xpra.log import debug_enabled_categories
    debug = list(SUBPROCESS_DEBUG)
    for f in ("audio", "gstreamer"):
        if (DEBUG_SOUND or f in debug_enabled_categories) and (f not in debug):
            debug.append(f)
    if debug:
        # forward debug flags:
        command.append("--debug")
        command.append(",".join(debug))


class AudioSubprocessWrapper(SubprocessCaller):
    """ This utility superclass deals with the caller side of the audio subprocess wrapper:
        * starting the wrapper subprocess
        * handling state-changed signal, so we have a local copy of the current value ready
        * handle "info" packets, so we have a cached copy
        * forward get/set volume calls (get_volume uses the value found in "info")
    """

    def __init__(self, description: str):
        super().__init__(description)
        self.state = "stopped"
        self.codec = "unknown"
        self.codec_description = ""
        self.info = {}
        # hook some default packet handlers:
        self.connect("state-changed", self.state_changed)
        self.connect("info", self.info_update)
        self.connect("signal", self.subprocess_signal)

    def get_env(self) -> dict[str, str]:
        env = super().get_env()
        env.update(get_audio_wrapper_env())
        env.pop("DISPLAY", None)
        # env.pop("WAYLAND_DISPLAY", None)
        return env

    def start(self) -> None:
        self.state = "starting"
        super().start()
        log("start() %s subprocess(%s)=%s", self.description, self.command, self.process.pid)
        GLib.timeout_add(SOUND_START_TIMEOUT, self.verify_started)

    def cleanup(self) -> None:
        log("cleanup() sending cleanup request to %s", self.description)
        self.send("cleanup")
        # cleanup should cause the process to exit
        GLib.timeout_add(500, self.send, "stop")
        GLib.timeout_add(1000, self.send, "exit")
        GLib.timeout_add(1500, self.stop)

    def verify_started(self) -> None:
        p = self.process
        log("verify_started() process=%s, info=%s, codec=%s", p, self.info, self.codec)
        if p is None:
            log("no process")
            return
        if p.poll() is not None:
            log("process has already terminated: exit code=%s", p.poll())
            return
        # if we don't get an "info" packet, then the pipeline must have failed to start
        if not self.info:
            log.warn("Warning: the %s process has failed to start", self.description)
            self.cleanup()

    def subprocess_signal(self, _wrapper, proc) -> None:
        log("subprocess_signal: %s", proc)
        # call via idle_add to prevent deadlocks on win32!
        GLib.idle_add(self.stop_protocol)

    def state_changed(self, _wrapper, new_state: str) -> None:
        self.state = new_state

    def get_state(self) -> str:
        return self.state

    def get_info(self) -> dict:
        return self.info

    def info_update(self, _wrapper, info: dict) -> None:
        log("info_update: %s", info)
        self.info.update(info)
        self.info["time"] = int(monotonic())
        p = self.process
        if p and not p.poll():
            self.info["pid"] = p.pid
        self.codec_description = info.get("codec_description")

    def set_volume(self, v: float) -> None:
        self.send("set_volume", int(v * 100))

    def get_volume(self) -> float:
        return self.info.get("volume", 100) / 100.0


class SourceSubprocessWrapper(AudioSubprocessWrapper):

    def __init__(self, plugin, _options, codecs, volume, element_options):
        super().__init__("audio capture")
        self.large_packets = ["new-buffer"]
        self.command = get_full_audio_command() + [
            "_audio_record", "-", "-",
            plugin or "", format_element_options(element_options),
            ",".join(codecs), "",
            str(volume),
        ]
        _add_debug_args(self.command)

    def __repr__(self):
        proc = self.process
        if proc:
            try:
                return "source_subprocess_wrapper(%s)" % proc.pid
            except AttributeError:
                pass
        return f"source_subprocess_wrapper({proc})"


class SinkSubprocessWrapper(AudioSubprocessWrapper):

    def __init__(self, plugin, codec, volume, element_options):
        super().__init__("audio playback")
        self.large_packets = ["add_data"]
        self.codec = codec
        self.command = get_full_audio_command() + [
            "_audio_play", "-", "-",
            plugin or "", format_element_options(element_options),
            codec, "",
            str(volume),
        ]
        _add_debug_args(self.command)

    def add_data(self, data: bytes, metadata: dict, packet_metadata=()) -> None:
        if DEBUG_SOUND:
            log("add_data(%s bytes, %s, %s) forwarding to %s", len(data), metadata, len(packet_metadata), self.protocol)
        self.send("add_data", data, metadata, packet_metadata)

    def __repr__(self):
        proc = self.process
        if proc:
            try:
                return "sink_subprocess_wrapper(%s)" % proc.pid
            except AttributeError:
                pass
        return "sink_subprocess_wrapper(%s)" % proc


def start_sending_audio(plugins, audio_source_plugin: str, device: str, codec: str, volume: float,
                        want_monitor_device: bool,
                        remote_decoders: Iterable[str],
                        remote_pulseaudio_server: str, remote_pulseaudio_id: str
                        ) -> SourceSubprocessWrapper | None:
    log("start_sending_audio%s",
        (plugins, audio_source_plugin, device, codec, volume, want_monitor_device,
         remote_decoders, remote_pulseaudio_server, remote_pulseaudio_id))
    with log.trap_error("Error setting up %r audio source" % (audio_source_plugin or "auto")):
        # info about the remote end:
        PAInfo = namedtuple("PAInfo", ("pulseaudio_server", "pulseaudio_id", "remote_decoders"))
        remote = PAInfo(pulseaudio_server=remote_pulseaudio_server,
                        pulseaudio_id=remote_pulseaudio_id,
                        remote_decoders=remote_decoders)
        plugin, options = parse_audio_source(plugins, audio_source_plugin, device, want_monitor_device, remote)
        if plugin:
            log("parsed '%s':", audio_source_plugin)
            log("plugin=%s", plugin)
            log("options=%s", options)
            return SourceSubprocessWrapper(plugin, options, remote_decoders, volume, options)
    return None


def start_receiving_audio(codec: str) -> SinkSubprocessWrapper:
    log("start_receiving_audio(%s)", codec)
    with log.trap_error("Error starting audio sink"):
        return SinkSubprocessWrapper(None, codec, 1.0, {})


def query_audio() -> typedict:
    import subprocess
    command = get_full_audio_command() + ["_audio_query"]
    _add_debug_args(command)
    kwargs = exec_kwargs()
    env = exec_env()
    env.update(get_audio_wrapper_env())
    env.pop("DISPLAY", None)
    log(f"query_audio() command={command!r}, env={env}, kwargs={kwargs}")
    proc = subprocess.Popen(command,
                            stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=env, universal_newlines=True,
                            **kwargs)
    with proc:
        out, err = proc.communicate(None)
    log(f"query_audio() process returned {proc.returncode}")
    log(f"query_audio() out={out!r}, err={err!r}")
    if proc.returncode != 0:
        return typedict()
    d = typedict()
    for x in out.splitlines():
        kv = x.split("=", 1)
        if len(kv) == 2:
            # ie: kv = ["decoders", "mp3,vorbis"]
            k, v = kv
            # fugly warning: all the other values are lists.. but this one is not:
            if k != "python.bits":
                v = v.split(",")
            # d["decoders"] = ["mp3", "vorbis"]
            d[k] = v
    log(f"query_audio()={d}")
    return d



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.x11.bindings.xlib cimport Display

cdef dict get_monitor_properties(Display *display)
#cdef get_crtc_info(Display *display, XRRScreenResources *rsc, RRCrtc crtc)
#cdef get_output_properties(Display *display, RROutput output)
#cdef get_output_info(Display *display, XRRScreenResources *rsc, RROutput output)
#cdef get_mode_info(XRRModeInfo *mi, with_sync : bool)

cdef dict get_all_screen_properties(Display *display)
